/*
 *  Copyright 2020-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *  @file fbtools-compat.c
 *
 *  Provides functions that make libfixbuf 2.x mimic libfixbuf 3.x.  Part of
 *  libfbtools.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */


#include "fbtools-compat.h"

#if !FIXBUF_CHECK_VERSION(2, 99, 0)


void
fbTemplateIterInit(
    fbTemplateIter_t    *iter,
    const fbTemplate_t  *tmpl)
{
    memset(&iter->field, 0, sizeof(iter->field));
    iter->field.tmpl = tmpl;
    iter->tmpl = tmpl;
    iter->pos = 0;
    iter->max_pos = fbTemplateCountElements(tmpl);
}

fbTemplateField_t *
fbTemplateIterNext(
    fbTemplateIter_t  *iter)
{
    const fbInfoElement_t *tmpl_ie;

    if (iter->pos >= iter->max_pos) {
        return NULL;
    }

    /* Get the offset for the next element by adding the length of the current
     * element.  When returning the first element, the length of the "current
     * field" is zero, so the offset is not changed. */
    if (fbTemplateFieldGetLen(&iter->field) != FB_IE_VARLEN) {
        iter->field.offset += fbTemplateFieldGetLen(&(iter->field));
    } else {
        switch (fbTemplateFieldGetType(&iter->field)) {
          case FB_BASIC_LIST:
            iter->field.offset += sizeof(fbBasicList_t);
            break;
          case FB_SUB_TMPL_LIST:
            iter->field.offset += sizeof(fbSubTemplateList_t);
            break;
          case FB_SUB_TMPL_MULTI_LIST:
            iter->field.offset += sizeof(fbSubTemplateMultiList_t);
            break;
          case FB_OCTET_ARRAY:
          case FB_STRING:
            iter->field.offset += sizeof(fbVarfield_t);
            break;
          default:
            g_error("Invalid type for VARLEN field %u",
                    fbTemplateFieldGetType(&iter->field));
        }
    }

    tmpl_ie = fbTemplateGetIndexedIE((fbTemplate_t *)iter->tmpl, iter->pos);
    iter->field.ie = tmpl_ie->ref.canon;
    iter->field.midx = tmpl_ie->midx;
    iter->field.len = tmpl_ie->len;

    ++iter->pos;
    return &iter->field;
}

void *
fbRecordValueTakeVarfieldBuf(
    fbRecordValue_t  *value)
{
    void *buf = g_string_free(value->stringbuf, FALSE);
    value->stringbuf = NULL;
    return buf;
}


static void
fbRecordFillValue(
    const fbTemplateField_t  *field,
    const uint8_t            *buf,
    fbRecordValue_t          *value)
{
    memset(&value->v, 0, sizeof(value->v));
    value->ie = fbTemplateFieldGetIE(field);
    switch (fbTemplateFieldGetType(field)) {
      case FB_BOOL:
        /* According to spec: TRUE == 1, FALSE == 2 */
        value->v.u64 = (1 == *buf);
        break;
      case FB_UINT_8:
        value->v.u64 = *buf;
        break;
      case FB_UINT_16:
      case FB_UINT_32:
      case FB_UINT_64:
#if G_BYTE_ORDER == G_BIG_ENDIAN
        memcpy(((uint8_t *)&value->v.u64
                + sizeof(value->v.64) - fbTemplateFieldGetLen(field)),
               buf, fbTemplateFieldGetLen(field));
#else
        memcpy((uint8_t *)&value->v.u64, buf, fbTemplateFieldGetLen(field));
#endif  /* G_BYTE_ORDER == G_BIG_ENDIAN */
        break;
      case FB_IP4_ADDR:
        memcpy(&value->v.ip4, buf, sizeof(value->v.ip4));
        break;
      case FB_IP6_ADDR:
        memcpy(value->v.ip6, buf, sizeof(value->v.ip6));
        break;
      case FB_MAC_ADDR:
        memcpy(value->v.mac, buf, sizeof(value->v.mac));
        break;
      case FB_DT_SEC:
#if G_BYTE_ORDER == G_BIG_ENDIAN
        memcpy(((uint8_t *)&value->v.u64
                + sizeof(value->v.64) - fbTemplateFieldGetLen(field)),
               buf, fbTemplateFieldGetLen(field));
#else
        memcpy((uint8_t *)&value->v.u64, buf, fbTemplateFieldGetLen(field));
#endif  /* G_BYTE_ORDER == G_BIG_ENDIAN */
        break;
      case FB_DT_MILSEC:
        {
            imaxdiv_t   div;
            uint64_t   *u64;
#ifdef HAVE_ALIGNED_ACCESS_REQUIRED
            uint64_t    u64_local;
            memcpy(&u64_local, buf, sizeof(u64_local));
            u64 = &u64_local;
#else
            u64 = (uint64_t *)buf;
#endif  /* HAVE_ALIGNED_ACCESS_REQUIRED */
            div = imaxdiv(*u64, 1000);
            value->v.dt.tv_sec = div.quot;
            value->v.dt.tv_nsec = div.rem * 1000000;
        }
        break;
      case FB_DT_MICROSEC:
      case FB_DT_NANOSEC:
        {
            /* The number of seconds between the NTP epoch (Jan 1, 1900) and
             * the UNIX epoch (Jan 1, 1970).  Seventy 365-day years plus 17
             * leap days, at 86400 sec/day: ((70 * 365 + 17) * 86400) */
            const uint64_t NTP_EPOCH_TO_UNIX_EPOCH = UINT64_C(0x83AA7E80);
            /* When the NTP value rolls over in 2036, must add 2^32 seconds to
             * the UNIX time */
            const uint64_t NTP_ROLLOVER = UINT64_C(0x100000000);
            /* To get a proper value when converting to fractional seconds,
             * add 1<<31 before the >>32 to round up values.  Not doing so
             * introduces errors that can accumulate with repeated
             * conversions. */
            const uint64_t ROUNDING_DIFF = UINT64_C(0x80000000);
            /* Fractional mask to apply to NTP to get Nanoseconds */
            const uint64_t NANO_MASK = UINT64_C(0xffffffff);
            /* Fractional mask to use for Micoseconds -- ignore 11 LSB */
            const uint64_t MICRO_MASK = UINT64_C(0xfffff800);
            /* Nanoseconds per whole second (1e9) */
            const uint64_t NANO_PER_SEC = UINT64_C(1000000000);
            /* Microseconds per whole second (1e6) */
            const uint64_t MICRO_PER_SEC = UINT64_C(1000000);

            const uint64_t *u64;
#ifdef HAVE_ALIGNED_ACCESS_REQUIRED
            uint64_t        u64_local;

            memcpy(&u64_local, buf, sizeof(u64_local));
            u64 = &u64_local;
#else
            u64 = (const uint64_t *)buf;
#endif

            /* Right shift `*u64` by 32 to get the whole seconds since 1900.
             * Subtract the difference between the epochs to get a UNIX
             * time.
             *
             * Use the highest bit of *u64 to determine (assume) the NTP Era
             * and add NTP_ROLLOVER if Era 1; this is valid from 1968 to
             * 2104. */
            if ((*u64) >> 63) {
                /* Assume NTP Era 0 */
                /* valid for 1968-01-20 03:14:08Z to 2036-02-07 06:28:15Z */
                value->v.dt.tv_sec =
                    (int64_t)((*u64) >> 32) - NTP_EPOCH_TO_UNIX_EPOCH;
            } else {
                /* Assume NTP Era 1 */
                /* valid for 2036-02-07 06:28:16Z to 2104-02-26 09:42:23Z */
                value->v.dt.tv_sec = (int64_t)((*u64) >> 32) + NTP_ROLLOVER
                    - NTP_EPOCH_TO_UNIX_EPOCH;
            }

            /* Apply the appropriate mask to `*u64` to get the fractional
             * seconds.  Divide by 2^32 to get a floating point number that is
             * a fraction of a second, and multiply by units per second, but
             * do those in reverse order and use shift for the division. */
            if (fbTemplateFieldGetType(field) == FB_DT_MICROSEC) {
                /* Compute the number of microseconds and multiply by 1000 to
                 * get nsec ensuring a whole number of usec. */
                value->v.dt.tv_nsec =
                    ((((*u64) & MICRO_MASK) * MICRO_PER_SEC + ROUNDING_DIFF)
                     >> 32) * 1000;
            } else {
                value->v.dt.tv_nsec =
                    ((((*u64) & NANO_MASK) * NANO_PER_SEC + ROUNDING_DIFF)
                     >> 32);
            }
        }
        break;
      case FB_FLOAT_64:
        if (8 == fbTemplateFieldGetLen(field)) {
            memcpy(&value->v.dbl, buf, sizeof(value->v.dbl));
            break;
        }
        /* FALLTHROUGH */
      case FB_FLOAT_32:
        {
            float f;
            memcpy(&f, buf, sizeof(f));
            value->v.dbl = f;
            break;
        }
      case FB_INT_8:
        value->v.s64 = *buf;
        break;
      case FB_INT_16:
      case FB_INT_32:
      case FB_INT_64:
#if G_BYTE_ORDER == G_BIG_ENDIAN
        if (0x80u & *buf) {
            value->v.s64 = UINT64_MAX;
        }
        memcpy(((uint8_t *)&value->v.s64
                + sizeof(value->v.64) - fbTemplateFieldGetLen(field),
                buf, fbTemplateFieldGetLen(field));
#else  /* G_BYTE_ORDER == G_BIG_ENDIAN */
        if (0x80u & *(buf + fbTemplateFieldGetLen(field) - 1)) {
            value->v.s64 = UINT64_MAX;
        }
        memcpy((uint8_t *)&value->v.s64, buf, fbTemplateFieldGetLen(field));
#endif  /* G_BYTE_ORDER == G_BIG_ENDIAN */
        break;
      case FB_STRING:
      case FB_OCTET_ARRAY:
        {
            const char *vf_buf;
            size_t      vf_len;
            if (fbTemplateFieldGetLen(field) == FB_IE_VARLEN) {
                vf_buf = (char *)( ((fbVarfield_t *)buf)->buf);
                vf_len = ((fbVarfield_t *)buf)->len;
            } else {
                vf_buf = (char *)buf;
                vf_len = fbTemplateFieldGetLen(field);
            }
            if (NULL == value->stringbuf) {
                value->stringbuf = g_string_new_len(vf_buf, vf_len);
            } else {
                g_string_truncate(value->stringbuf, 0);
                g_string_append_len(value->stringbuf, vf_buf, vf_len);
            }
            value->v.varfield.buf = (uint8_t *)value->stringbuf->str;
            value->v.varfield.len = value->stringbuf->len;
        }
        break;
      case FB_BASIC_LIST:
        value->v.bl = (fbBasicList_t *)buf;
        break;
      case FB_SUB_TMPL_LIST:
        value->v.stl = (fbSubTemplateList_t *)buf;
        break;
      case FB_SUB_TMPL_MULTI_LIST:
        value->v.stml = (fbSubTemplateMultiList_t *)buf;
        break;
    }
}


gboolean
fbRecordGetValueForField(
    const fbRecord_t        *record,
    const fbTemplateField_t *field,
    fbRecordValue_t         *value)
{
    fbRecordFillValue(field, record->rec + field->offset, value);
    return TRUE;
}


gboolean
fBufNextRecord(
    fBuf_t      *fbuf,
    fbRecord_t  *record,
    GError     **err)
{
    if (!fBufSetInternalTemplate(fbuf, record->tid, err)) {
        return FALSE;
    }
    record->tmpl = fbSessionGetTemplate(
        fBufGetSession(fbuf), TRUE, record->tid, NULL);
    record->recsize = record->reccapacity;
    return fBufNext(fbuf, record->rec, &record->recsize, err);
}


uint16_t
fbBasicListGetElementLength(
    const fbBasicList_t  *bl)
{
    if (0 == fbBasicListCountElements(bl)
        || FB_IE_VARLEN == fbInfoElementGetLen(fbBasicListGetInfoElement(bl)))
    {
        return fbInfoElementGetLen(fbBasicListGetInfoElement(bl));
    }

    return (bl->dataLength / bl->numElements);
}


fbTemplateField_t *
fbBasicListGetTemplateField(
    const fbBasicList_t  *bl)
{
    static fbTemplateField_t field = FB_TEMPLATEFIELD_INIT;

    field.ie = fbBasicListGetInfoElement(bl);
    field.len = fbBasicListGetElementLength(bl);

    return &field;
}


gboolean
fbBasicListGetIndexedRecordValue(
    const fbBasicList_t  *bl,
    uint16_t              pos,
    fbRecordValue_t      *value)
{
    fbTemplateField_t  field = FB_TEMPLATEFIELD_INIT;
    const void        *item;

    item = fbBasicListGetIndexedDataPtr(bl, pos);
    if (item) {
        field.ie = fbBasicListGetInfoElement(bl);
        field.len = fbBasicListGetElementLength(bl);
        fbRecordFillValue(&field, item, value);
        return TRUE;
    }

    return FALSE;
}

#endif  /* #if !FIXBUF_CHECK_VERSION(2, 99, 0) */
