/*
 *  Copyright 2020-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *  @file fbtools-compat.h
 *
 *  Provides macros, typedefs, and function declarations that make libfixbuf
 *  2.x mimic libfixbuf 3.x.  Part of libfbtools.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#ifndef _FBTOOLS_COMPAT_H_
#define _FBTOOLS_COMPAT_H_
#include <fixbuf/public.h>

#if !defined(FIXBUF_CHECK_VERSION)

/* if the macro is not defined, it must be older than fixbuf 1.8.0.  always
 * return false */
#define FIXBUF_CHECK_VERSION(x,y,z)     0

#elif !defined(FIXBUF_VERSION_MAJOR)

/* fixbuf 1.8.0 defines FIXBUF_CHECK_VERSION() in public.h but does
 * not include the file that defines the variables the macro uses */
#include <fixbuf/version.h>

#endif  /* FIXBUF_CHECK_VERSION */

#if !FIXBUF_CHECK_VERSION(2, 99, 0)

#if !FIXBUF_CHECK_VERSION(2, 3, 0)
#error "This version of libfixbuf is too old.  v2.3.0 or later is required"
#endif


/* Define items added in libfixbuf-3.0.0 */



/* *****  Remove const from the first argument  ***** */

#define fBufListFree(template_, u8array_)                       \
    fBufListFree((fbTemplate_t *)(template_), (u8array_))

#define fbBasicListGetIndexedDataPtr(b_, p_)                    \
    fbBasicListGetIndexedDataPtr((fbBasicList_t *)(b_), (p_))

#define fbBasicListGetInfoElement(x_)                   \
    fbBasicListGetInfoElement((fbBasicList_t *)(x_))

#define fbBasicListGetSemantic(x_)                      \
    fbBasicListGetSemantic((fbBasicList_t *)(x_))

#define fbSessionGetDomain(s_)                  \
    fbSessionGetDomain((fbSession_t *)(s_))

#define fbSubTemplateListGetSemantic(x_)                        \
    fbSubTemplateListGetSemantic((fbSubTemplateList_t *)(x_))

#define fbSubTemplateListGetTemplate(x_)                        \
    fbSubTemplateListGetTemplate((fbSubTemplateList_t *)(x_))

#define fbSubTemplateListGetTemplateID(x_)                      \
    fbSubTemplateListGetTemplateID((fbSubTemplateList_t *)(x_))

#define fbSubTemplateMultiListGetSemantic(x_)                           \
    fbSubTemplateMultiListGetSemantic((fbSubTemplateMultiList_t *)(x_))

#define fbSubTemplateMultiListEntryGetTemplate(x_)      \
    fbSubTemplateMultiListEntryGetTemplate(             \
        (fbSubTemplateMultiListEntry_t *)(x_))

#define fbSubTemplateMultiListEntryGetTemplateID(x_)    \
    fbSubTemplateMultiListEntryGetTemplateID(           \
        (fbSubTemplateMultiListEntry_t *)(x_))

#define fbTemplateCountElements(x_)                     \
    fbTemplateCountElements((fbTemplate_t *)(x_))

#define fbTemplateGetContext(x_)                \
    fbTemplateGetContext((fbTemplate_t *)(x_))



/* *****  Convenience Macros and Small Functions  ***** */


/* ***  fbBasicList_t  *** */

#define fBufSetAutomaticNextMessage(fb, automatic)      \
    fBufSetAutomaticMode((fb), (automatic))


/* ***  fbBasicList_t  *** */

#define fbBLNext(type, basicList, current)                      \
    ((type *)(fbBasicListGetNextPtr(                            \
                  (fbBasicList_t *)(basicList), (current))))


/* ***  fbInfoElement_t  *** */

#define fbInfoElementGetLen(infoElement)        (infoElement)->len
#define fbInfoElementGetPEN(infoElement)        (infoElement)->ent
#define fbInfoElementGetType(infoElement)       (infoElement)->type
#define fbInfoElementGetUnits(infoElement)      \
    (FB_IE_UNITS((infoElement)->flags))

#define fbInfoElementIsList(infoElement)                \
    ((infoElement)->type == FB_BASIC_LIST               \
     || (infoElement)->type == FB_SUB_TMPL_LIST         \
     || (infoElement)->type == FB_SUB_TMPL_MULTI_LIST)


/* ***  fbSession_t  *** */

#define fbSessionAddTemplate(_a, _b, _c, _d, _e, _f)        \
    fbSessionAddTemplate(_a, _b, _c, _d, _f)


/* ***  fbSubTemplateList_t  *** */

#define fbSTLNext(type, subTemplateList, current)                       \
    ((type *)(fbSubTemplateListGetNextPtr(                              \
                  (fbSubTemplateList_t *)(subTemplateList), (current))))


/* ***  fbSubTemplateMultiList_t and fbSubTemplateMultiListEntry_t  *** */

#define fbSTMLNext(subTemplateMultiList, current)                       \
    fbSubTemplateMultiListGetNextEntry(                                 \
        (fbSubTemplateMultiList_t *)(subTemplateMultiList), (current))

#define fbSTMLEntryNext(type, subTemplateMultiListEntry, current) \
    ((type *) \
     (fbSubTemplateMultiListEntryNextDataPtr(                           \
         (fbSubTemplateMultiListEntry_t *)(subTemplateMultiListEntry),  \
         (current))))


/* ***  fbTemplate_t  *** */

#define FB_TMPL_IS_META_ELEMENT     (1u << 1)

#define fbTemplateIsMetadata(template_, flags_)                 \
    ((FB_TMPL_IS_META_ELEMENT == (flags_))                      \
     ? fbInfoModelTypeInfoRecord((fbTemplate_t *)(template_))   \
     : 0)



/* *****  New Types and Functions  ***** */

/* ***  fbTemplateField_t  *** */

/*  fbTemplateField_t is an fbInfoElement_t when part of an fbTemplate_t */

typedef struct fbTemplateField_st {
    const fbInfoElement_t   *ie;
    const fbTemplate_t      *tmpl;
    uint16_t                 offset;
    uint16_t                 len;
    uint16_t                 midx;
} fbTemplateField_t;

#define FB_TEMPLATEFIELD_INIT   {NULL, NULL, 0, 0, 0}

#define fbTemplateFieldCheckIdent(tfield_, _pen, _id)                   \
    ((_pen) == (tfield_)->ie->ent && (_id) == (tfield_)->ie->num)

#define fbTemplateFieldGetIE(tfield_)     (tfield_)->ie
#define fbTemplateFieldGetId(tfield_)     (tfield_)->ie->num
#define fbTemplateFieldGetLen(tfield_)    (tfield_)->len
#define fbTemplateFieldGetName(tfield_)   (tfield_)->ie->ref.name
#define fbTemplateFieldGetOffset(tfield_) (tfield_)->offset
#define fbTemplateFieldGetPEN(tfield_)    (tfield_)->ie->ent
#define fbTemplateFieldGetRepeat(tfield_) (tfield_)->midx
#define fbTemplateFieldGetType(tfield_)   (tfield_)->ie->type


/* ***  fbTemplateIter_t  *** */

/*  fbTemplateIter_t loops over the entries in an fbTemplate_t */

typedef struct fbTemplateIter_st {
    const fbTemplate_t  *tmpl;
    fbTemplateField_t    field;
    uint16_t             max_pos;
    uint16_t             pos;
} fbTemplateIter_t;

#define FB_TEMPLATE_ITER_NULL  {NULL, FB_TEMPLATEFIELD_INIT, 0, 0}


/*  Initializes the iterator for looping over an fbTemplate_t. */
void
fbTemplateIterInit(
    fbTemplateIter_t    *iter,
    const fbTemplate_t  *tmpl);

/*  Gets the first/next field in the fbTemplate_t the iterator is bound to.
 *  Returns NULL when all entries have been visited. */
fbTemplateField_t *
fbTemplateIterNext(
    fbTemplateIter_t  *iter);

/*
 *  Gets the position (relative to the number of elements) of the iterator,
 *  where 0 means the first entry in the template and
 *  fbTemplateCountElements()-1 is the final entry.
 *
 *  Returns UINT16_MAX if fbTemplateIterNext() has not been called; returns
 *  fbTemplateCountElements() if the iterator is exhausted.
 */
#define fbTemplateIterGetPosition(t_iter_)                      \
    ((0 == (t_iter_)->pos) ? UINT16_MAX : ((t_iter_)->pos - 1))


/* ***  fbRecord_t  *** */

/*  fbRecord_t holds the data for an IPFIX record and a pointer to its
 *  fbTemplate_t */

typedef struct fbRecord_st {
    /** The template that describes the bytes in `rec`. */
    const fbTemplate_t  *tmpl;
    /** The buffer holding the data for a record. */
    uint8_t             *rec;
    /** The number of bytes allocated for the `rec` buffer. */
    size_t               reccapacity;
    /** The number of bytes used by the record current in `rec`. */
    size_t               recsize;
    /** The template ID for `tmpl`. */
    uint16_t             tid;
} fbRecord_t;

#define FB_RECORD_INIT {NULL, NULL, 0, 0, 0}

/*  Recursively frees all memory held by lists in an fbRecord_t */
#define fbRecordFreeLists(record_)                      \
    fBufListFree((record_)->tmpl, (record_)->rec)



/* ***  fbRecordValue_t  *** */

/*  fbRecordValue_t holds one value when requested from an fbRecord_t */

typedef struct fbRecordValue_st {
    /** The element that describes this data. */
    const fbInfoElement_t  *ie;
    /** An internal buffer used to store the `varfield` value. */
    GString                *stringbuf;
    /** The value depending on the type of `ie`. */
    union v_un {
        /**
         * Used when `ie` is @ref FB_IP6_ADDR.
         */
        uint8_t                          ip6[16];
        /**
         * Used when `ie` is @ref FB_BASIC_LIST.
         */
        const fbBasicList_t             *bl;
        /**
         * Used when `ie` is @ref FB_SUB_TMPL_LIST.
         */
        const fbSubTemplateList_t       *stl;
        /**
         * Used when `ie` is @ref FB_SUB_TMPL_MULTI_LIST.
         */
        const fbSubTemplateMultiList_t  *stml;
        /**
         * Used when `ie` is @ref FB_STRING or @ref FB_OCTET_ARRAY.  The `buf`
         * member of `varfield` is backed by a GString; the value is
         * terminated by `\0`.  The `len` member is the length of `buf` not
         * including the terminator.
         */
        fbVarfield_t                     varfield;
        /**
         * Used when `ie` is @ref FB_DT_SEC, @ref FB_DT_MILSEC, @ref
         * FB_DT_MICROSEC, or @ref FB_DT_NANOSEC.  `dt.tv_nsec` is in
         * nanoseconds; divide by 1000000 to get milliseconds or 1000 to get
         * microseconds.
         */
        struct timespec                  dt;
        /**
         * Used when `ie` is @ref FB_IP4_ADDR.
         */
        uint32_t                         ip4;
        /**
         * Used when `ie` is @ref FB_UINT_8, @ref FB_UINT_16, @ref FB_UINT_32,
         * @ref FB_UINT_64, or @ref FB_BOOL (0 == false, 1 == true).
         */
        uint64_t                         u64;
        /**
         * Used when `ie` is @ref FB_INT_8, @ref FB_INT_16, @ref FB_INT_32, or
         * @ref FB_INT_64.
         */
        int64_t                          s64;
        /**
         * Used when `ie` is @ref FB_FLOAT_64 or @ref FB_FLOAT_32.
         */
        double                           dbl;
        /**
         * Used when `ie` is @ref FB_MAC_ADDR.
         */
        uint8_t                          mac[6];
    } v;
} fbRecordValue_t;

#define FB_RECORD_VALUE_INIT \
    { NULL, NULL, { {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} } }


/*  Takes ownership of the string held by the `stringbuf` member of an
 *  fbRecordValue_t and sets the `stringbuf` member to NULL. */
void *
fbRecordValueTakeVarfieldBuf(
    fbRecordValue_t  *value);

/*  Fills `value` with the value for `field` in `record`. */
gboolean
fbRecordGetValueForField(
    const fbRecord_t        *record,
    const fbTemplateField_t *field,
    fbRecordValue_t         *value);

/*  Frees the `stringbuf` member of an fbRecordValue_t. */
#define fbRecordValueClear(recval_)                     \
    do {                                                \
        g_string_free((recval_)->stringbuf, TRUE);      \
        memset((recval_), 0, sizeof(*(recval_)));       \
    } while(0)

/*
 *  Reads data from the buffer `fbuf` and fills the `rec` member of `record`
 *  using the internal template specified by the `tid` member of `record`.
 */
gboolean
fBufNextRecord(
    fBuf_t      *fbuf,
    fbRecord_t  *record,
    GError     **err);



/* ***  fbBasicList_t  *** */

/*  Returns the length of the element in the fbBasicList_t. */
uint16_t
fbBasicListGetElementLength(
    const fbBasicList_t  *bl);

/*  Returns a simulated fbTemplateField_t for an fbBasicList_t. */
fbTemplateField_t *
fbBasicListGetTemplateField(
    const fbBasicList_t  *bl);


/*  Fills `value` with the data from the `pos` entry in the fbBasicList_t
 *  `bl`. */
gboolean
fbBasicListGetIndexedRecordValue(
    const fbBasicList_t  *bl,
    uint16_t              pos,
    fbRecordValue_t      *value);


#endif  /* #if !FIXBUF_CHECK_VERSION(2, 99, 0) */
#endif  /* #ifndef _FBTOOLS_COMPAT_H_ */
