/*
 *  Copyright 2014-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *    @file fbtools-lib.c
 *
 *    Helper functions; part of libfbtools.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#include "fbtools-lib.h"


/* User options for input, output, and XML cert element files. */
static struct fbtools_user_opts_st {
 gchar      **cert_element_file_paths;
 gchar      **element_files;
 char        *inspec;
 char        *outspec;
 gboolean     no_cert_elements;
 gboolean     rfc5610;
} fbtools_user_opts;

/* The IPFIX Info Model */
static fbInfoModel_t *model = NULL;

/* Maps from a template id to its name. */
static GHashTable         *template_names;

/* Any array holding all fbToolsTemplateName_t objects that have been
 * created. */
static GPtrArray *t_name_ctx;

/* This is created to hold the locations of the template-id and template-name
 * elements within the template-name template.  The caller is expected to add
 * this to the template callback.
 */
typedef struct fbToolsTemplateName_st fbToolsTemplateName_t;
struct fbToolsTemplateName_st {
    fbTemplateField_t  templateId;
    fbTemplateField_t  templateName;
};

typedef struct fbToolsRfc5610_st fbToolsRfc5610_t;

struct fbToolsRfc5610_st {
};



/* map from --octet-format and --string-format argument to setting */
static struct formatNameValueMap_st {
    const char         *name;
    fbToolsFormat_t  value;
    /* Whether this option is allowed for octet_array types only (1), for
     * string types only (2), or both (3) */
    unsigned int        octets_string;
} formatNameValueMap[] = {
    {"ascii",       FBTOOLS_FMT_ASCII,       3},
    {"base64",      FBTOOLS_FMT_BASE64,      3},
    {"empty",       FBTOOLS_FMT_EMPTY,       3},
    {"hexadecimal", FBTOOLS_FMT_HEXADECIMAL, 3},
    {"hexdump",     FBTOOLS_FMT_HEXDUMP,     3},
    {"octets",      FBTOOLS_FMT_OCTETS,      3},
    {"none",        FBTOOLS_FMT_EMPTY,       3},
    {"quoted",      FBTOOLS_FMT_QUOTED,      2},
    {"raw",         FBTOOLS_FMT_RAW,         2},
    {"short-hex",   FBTOOLS_FMT_SHORTHEX,    1},
    {"string",      FBTOOLS_FMT_STRING,      3},
    {"unicode",     FBTOOLS_FMT_UNICODE,     3},
    {NULL,          FBTOOLS_FMT_EMPTY,       0}
};


static GOptionEntry io_xml_options[] = {
    {"in", 'i', 0, G_OPTION_ARG_STRING, &fbtools_user_opts.inspec,
     (WRAP_OPT "Specifies the IPFIX file to process [-]"), "path"},
    {"out", 'o', 0, G_OPTION_ARG_STRING, &fbtools_user_opts.outspec,
     (WRAP_OPT "Specifies the output file's location [-]"), "path"},
    {"cert-element-path", '\0', 0, G_OPTION_ARG_FILENAME_ARRAY,
     &fbtools_user_opts.cert_element_file_paths,
     (WRAP_OPT "Adds a search directory for the CERT NetSA IE XML file; takes"
      WRAP_OPT "precedence over default locations.  May be repeated, and first"
      WRAP_OPT "directory that holds a '" CERT_IPFIX_BASENAME "' file is used"),
     "directory"},
    {"element-file", 'e', 0, G_OPTION_ARG_FILENAME_ARRAY,
     &fbtools_user_opts.element_files,
     (WRAP_OPT "Loads information elements from the given XML file (in addition"
      WRAP_OPT "to the CERT NetSA elements). May be repeated"),
     "path"},
    {"no-cert-elements", '\0', 0, G_OPTION_ARG_NONE,
     &fbtools_user_opts.no_cert_elements,
     (WRAP_OPT "Does not load the XML file of CERT NetSA information elements"),
     NULL},
    {"rfc5610", '\0', 0, G_OPTION_ARG_NONE, &fbtools_user_opts.rfc5610,
     (WRAP_OPT "Adds IEs to the information model that are specified in the"
      WRAP_OPT "input stream as element type option records"),
     NULL},
    { NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
};

void
fbToolsAddOptionsIOXml(
    GOptionGroup   *group)
{
    g_option_group_add_entries(group, io_xml_options);
}



/**
 *    Prints the help text for the options and exits the problem.
 *
 *    This is a callback invoked by the options parser when --help is given.
 */
static void
fbToolsPrintHelp(
    const gchar  *option_name,
    const gchar  *option_arg,
    gpointer      v_context,
    GError      **error)
{
    GOptionContext *ctx;
    char *help;

    (void)option_name;
    (void)option_arg;
    (void)error;

    ctx = (GOptionContext *)v_context;
    help = g_option_context_get_help(ctx, FALSE, NULL);

    printf("%s", help);
    g_free(help);
    exit(0);
}

/*    Registers the --help option. */
void
fbToolsAddOptionsHelp(
    GOptionGroup   *group)
{
    GOptionEntry help_option[] = {
        {"help", 'h', G_OPTION_FLAG_NO_ARG,
         G_OPTION_ARG_CALLBACK, &fbToolsPrintHelp,
         WRAP_OPT "Shows this brief description of the options", NULL},
        { NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
    };

    g_option_group_add_entries(group, help_option);
}



/**
 *    Prints the version text for the program and exits.
 *
 *    This is a callback invoked by the options parser when --version is
 *    given.
 */
static void
fbToolsPrintVersion(
    const gchar  *option_name,
    const gchar  *option_arg,
    gpointer      data,
    GError      **error)
{
    (void)option_name;
    (void)option_arg;
    (void)data;
    (void)error;

    fprintf(stdout, "%s version %s "
            "(C) 2014-2023 Carnegie Mellon University.\n",
            g_get_prgname(), FBTOOLS_PACKAGE_VERISON);
    fprintf(stdout, "GNU General Public License (GPL) Rights "
            "pursuant to Version 2, June 1991\n");
    fprintf(stdout, "Some included library code covered by LGPL 2.1; "
            "see source for details.\n");
    fprintf(stdout, "Send bug reports, feature requests, and comments to "
            "netsa-help@cert.org.\n");
    exit(0);
}

/*    Registers the --version option. */
void
fbToolsAddOptionsVersion(
    GOptionGroup   *group)
{
    GOptionEntry version_option[] = {
        {"version", 'V', G_OPTION_FLAG_NO_ARG,
         G_OPTION_ARG_CALLBACK, &fbToolsPrintVersion,
         WRAP_OPT "Prints the application version to stdout and exits", NULL},
        { NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
    };

    g_option_group_add_entries(group, version_option);
}






/**
 * fbToolsPrintIP6Address
 *
 *
 */
void
fbToolsPrintIP6Address(
    FILE          *fp,
    const uint8_t *ipaddr)
{
    const uint8_t *aqp;
    uint16_t       aq;
    gboolean       colon_start = FALSE;
    gboolean       colon_end = FALSE;

    for (aqp = ipaddr; aqp < ipaddr + 16; aqp += 2) {
        aq = g_ntohs(*((const uint16_t *)aqp));
        if (aq || colon_end) {
            if (aqp < ipaddr + 14) {
                fprintf(fp, "%04hx:", aq);
            } else {
                fprintf(fp, "%04hx", aq);
            }
            if (colon_start) {
                colon_end = TRUE;
            }
        } else if (!colon_start) {
            if (aqp == ipaddr) {
                fprintf(fp, "::");
            } else {
                fprintf(fp, ":");
            }
            colon_start = TRUE;
        }
    }
}


/**
 *    Prints the fbVarfield_t as ASCII.
 *
 *    Specifically octets in the range 0x20 (space) through 0x7e (tilde) are
 *    printed as ASCII, where 0x22 (quotation mark) and 0x5C (reverse solidus)
 *    are represented as a two-character sequence ("\"" or "\\"), and all
 *    other octets are represented as a five-charcter sequence "\\xhh" where h
 *    is a single hexadecimal digit.
 */
void
fbToolsPrintAscii(
    FILE               *fp,
    const fbVarfield_t *vf)
{
    static const char ascii_chars[] =
        " !#$%&'()*+,-./0123456789:;<=>?"   /* 0x20-0x21,0x23-0x3f (no \") */
        "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_"   /* 0x40-0x5b,0x5d-0x5f (no \\) */
        "`abcdefghijklmnopqrstuvwxyz{|}~";  /* 0x60-0x7e (no del) */
    const uint8_t    *c;
    size_t            len;

    if (strspn((char *)vf->buf, ascii_chars) == vf->len) {
        fprintf(fp, "\"%s\"", vf->buf);
        return;
    }

    len = vf->len;
    c = vf->buf;

    fprintf(fp, "\"");
    for ( ; len > 0; --len, ++c) {
        if (*c < (uint8_t)' ' || *c > (uint8_t)'~') {
            fprintf(fp, "\\\\x%02x", *c);
        } else if (*c == (uint8_t)'\\') {
            fprintf(fp, "\\\\");
        } else if (*c == (uint8_t)'\"') {
            fprintf(fp, "\\\"");
        } else {
            fprintf(fp, "%c", *c);
        }
    }
    fprintf(fp, "\"");
}


/**
 *    Prints the fbVarfield_t as UTF-8 encoded Unicode.  The caller must check
 *    the that value is valid UTF-8 before calling this function.
 *
 *    Specifically, writes printable unicode characters to the output and
 *    using a two-character escape sequence for U+0022 (quotation mark, "\""),
 *    U+005C (reverse solidus, "\\"), U+0009 (tab, "\t"), U+000A (line feed,
 *    "\n"), and U+000D (carriage return, "\r").  Non-printable unicode
 *    characters in the range U+0001 to U+FFFF are represented as "\uHHHH" and
 *    those in the range U+10000 to U+10FFFF as the UTF-16 surrogate pair
 *    "\uHHHH\uHHHH".
 */
void
fbToolsPrintUnicode(
    FILE               *fp,
    const fbVarfield_t *vf)
{
    GString     *gstr;
    gunichar     p;
    const gchar *c;

    gstr = g_string_sized_new(4 * (1 + vf->len));

    for (c = (const gchar *)vf->buf;
         (p = g_utf8_get_char(c)) != '\0';
         c = g_utf8_next_char(c))
    {
        /* Could also encode U+0008 backspace as \b, U+000B vertical-tab as
         * \v, and U+000C form-feed/new-page as \f, though \v is not listed in
         * RFC8259 */
        switch (p) {
          case '\t':
            /* U+0009 tab */
            g_string_append(gstr, "\\t");
            break;
          case '\n':
            /* U+000A newline */
            g_string_append(gstr, "\\n");
            break;
          case '\r':
            /* U+000D carriage-return */
            g_string_append(gstr, "\\r");
            break;
          case '"':
            g_string_append(gstr, "\\\"");
            break;
          case '\\':
            g_string_append(gstr, "\\\\");
            break;
          default:
            if (g_unichar_isprint(p)) {
                g_string_append_unichar(gstr, p);
            } else if (p <= 0xFFFF) {
                g_string_append_printf(gstr, "\\u%04" PRIX32, p);
            } else if (p <= 0x10FFFF) {
                /* Encode as a UTF-16 surrogate pair */
                g_string_append_printf(gstr,
                                       "\\u%04" PRIX32 "\\u%04" PRIX32,
                                       (0xD800 | ((p - 0x10000) >> 10)),
                                       (0xDC00 | ((p - 0x10000) & 0x03FF)));
            }
            break;
        }
    }

    fprintf(fp, "\"%s\"", gstr->str);
    g_string_free(gstr, TRUE);
}






/**
 *    Gets the format from the argument to --string-format or --octet-format.
 *    Exits the program if the format is invalid.
 */
fbToolsFormat_t
fbToolsParseStringOctetFormat(
    fbToolsStringOctetType_t  arg_type,
    const char                  *arg_value,
    const char                  *option_name)
{
    const struct formatNameValueMap_st *fm = formatNameValueMap;

    do {
        if (0 == strcmp(arg_value, fm->name)) {
            if ((arg_type & fm->octets_string) != 0) {
                return fm->value;
            }
            /* Invalid for this option */
            break;
        }
        ++fm;
    } while (fm->name != NULL);

    fprintf(stderr, "%s: Unrecognized %s option: %s\n",
            g_get_prgname(), option_name, arg_value);
    exit(1);
}



/* Create map from TID to template name */
void
fbToolsTemplateNamesInit(
    void)
{
    template_names = g_hash_table_new_full(g_direct_hash, NULL, NULL, g_free);
    t_name_ctx = g_ptr_array_new_full(64, g_free);
}

/* Destroy map from TID to template name */
void
fbToolsTemplateNamesDestroy(
    void)
{
    g_hash_table_destroy(template_names);
    template_names = NULL;
    g_ptr_array_free(t_name_ctx, TRUE);
    t_name_ctx = NULL;
}

/* Return true if any template name records were seen */
gboolean
fbToolsTemplateNamesSeen(
    void)
{
    return (template_names != NULL
            && g_hash_table_size(template_names) > 0);
}

/* Return the template name given a TID or the empty string */
const char *
fbToolsTemplateNamesGetByTid(
    uint16_t  tid)
{
    static const char no_name[] = "";
    const void *name;

    name = g_hash_table_lookup(template_names, GINT_TO_POINTER(tid));
    return ((NULL != name) ? (const char *)name : no_name);
}

/**
 *  Examines the template describing a template name to find the position of
 *  each element of interest.  This is a helper for the new template callback.
 */
fbToolsTemplateName_t *
fbToolsTemplateNamesExamineTmpl(
    const fbTemplate_t *tmpl)
{
    fbTemplateIter_t iter = FB_TEMPLATE_ITER_NULL;
    const fbTemplateField_t *field;
    fbToolsTemplateName_t *ctx;

    ctx = g_new0(fbToolsTemplateName_t, 1);

    fbTemplateIterInit(&iter, tmpl);
    while ((field = fbTemplateIterNext(&iter)) != NULL) {
        if (NULL == ctx->templateId.tmpl
            && fbTemplateFieldCheckIdent(field, 0, 145))
        {
            ctx->templateId = *field;
            if (NULL != ctx->templateName.tmpl) {
                break;
            }
        } else if (NULL == ctx->templateName.tmpl
                   && fbTemplateFieldCheckIdent(field, CERT_PEN, 1000))
        {
            ctx->templateName = *field;
            if (NULL != ctx->templateId.tmpl) {
                break;
            }
        }
    }

    if (NULL == ctx->templateId.tmpl || NULL == ctx->templateName.tmpl) {
        g_free(ctx);
        return NULL;
    }

    g_ptr_array_add(t_name_ctx, (gpointer)ctx);
    return ctx;
}

/**
 *    Parse a template name options record and insert the TID/name
 *    pair into the template_names hash table.
 */
void
fbToolsTemplateNamesExamineRecord(
    const fbToolsTemplateName_t *ctx,
    const fbRecord_t               *record)
{
    fbRecordValue_t          value = FB_RECORD_VALUE_INIT;
    uint16_t tid;

    if (NULL == ctx) {
        return;
    }

    /* templateId */
    fbRecordGetValueForField(record, &ctx->templateId, &value);
    tid = value.v.u64;

    /* templateName */
    fbRecordGetValueForField(record, &ctx->templateName, &value);
    if (tid >= FB_TID_MIN_DATA && value.v.varfield.len > 0) {
        g_hash_table_replace(template_names, GINT_TO_POINTER(tid),
                             fbRecordValueTakeVarfieldBuf(&value));
    } else {
        fbRecordValueClear(&value);
    }
}



fbToolsRfc5610_t *
fbToolsRfc5610ExamineTmpl(
    const fbTemplate_t *tmpl)
{
    if (fbtools_user_opts.rfc5610
        && fbTemplateIsMetadata(tmpl, FB_TMPL_IS_META_ELEMENT))
    {
        return g_new0(fbToolsRfc5610_t, 1);
    }
    return NULL;
}



/**
 *    Process an RFC5610 Record describing an InfoElement and add that
 *    element to the InfoModel.
 */
gboolean
fbToolsRfc5610ExamineRecord(
    const fbToolsRfc5610_t  *ctx,
    const fbRecord_t           *record)
{
    fbTemplateIter_t         iter = FB_TEMPLATE_ITER_NULL;
    fbRecordValue_t          value = FB_RECORD_VALUE_INIT;
    fbInfoElementOptRec_t    rec;
    const fbTemplateField_t *field;
    gboolean                 rv;

    (void)ctx;

    memset(&rec, 0, sizeof(rec));

    fbTemplateIterInit(&iter, record->tmpl);
    while ((field = fbTemplateIterNext(&iter))) {
        fbRecordGetValueForField(record, field, &value);
        if (fbTemplateFieldGetPEN(field) == 0) {
            switch (fbTemplateFieldGetId(field)) {
              case 346:
                /* privateEnterpriseNumber */
                rec.ie_pen = value.v.u64;
                break;
              case 303:
                /* informationElementId */
                rec.ie_id = value.v.u64;
                break;
              case 339:
                /* informationElementDataType */
                rec.ie_type = value.v.u64;
                break;
              case 344:
                /* informationElementSemantics */
                rec.ie_semantic = value.v.u64;
                break;
              case 345:
                /* informationElementUnits */
                rec.ie_units = value.v.u64;
                break;
              case 342:
                /* informationElementRangeBegin */
                rec.ie_range_begin = value.v.u64;
                break;
              case 343:
                /* informationElementRangeEnd */
                rec.ie_range_end = value.v.u64;
                break;
              case 341:
                /* informationElementName */
                rec.ie_name.len = value.v.varfield.len;
                rec.ie_name.buf =
                    (uint8_t *)fbRecordValueTakeVarfieldBuf(&value);
                break;
              case 340:
                /* informationElementDescription */
                rec.ie_desc.len = value.v.varfield.len;
                rec.ie_desc.buf =
                    (uint8_t *)fbRecordValueTakeVarfieldBuf(&value);
                break;
            }
        }
    }

    rv = fbInfoElementAddOptRecElement(model, &rec);
    g_free(rec.ie_name.buf);
    g_free(rec.ie_desc.buf);
    return rv;
}





static gchar *
fbToolsCheckXml(
    const gchar *filepath)
{
    gboolean verbose = TRUE;

    if (!g_file_test(filepath, G_FILE_TEST_EXISTS)) {
        if (verbose) {
            g_debug("Checking for %s: Ignoring non-existent path '%s'",
                    CERT_IPFIX_BASENAME, filepath);
        }
        return NULL;
    }

    if (g_file_test(filepath, G_FILE_TEST_IS_REGULAR)) {
        return g_strdup(filepath);
    }

    if (g_file_test(filepath, G_FILE_TEST_IS_DIR)) {
        gchar *f = g_build_filename(filepath, CERT_IPFIX_BASENAME, NULL);
        if (g_file_test(f, G_FILE_TEST_EXISTS)) {
            return f;
        }
        g_free(f);
        if (verbose) {
            g_debug("Checking for %s: No such file in directory '%s'",
                    CERT_IPFIX_BASENAME, filepath);
        }
        return NULL;
    }

    if (verbose) {
        g_debug("Checking for %s: Ignoring non-file, non-directory path '%s'",
                CERT_IPFIX_BASENAME, filepath);
    }
    return NULL;
}



/**
 *    Attempts to find the cert_ipfix.xml file and returns its
 *    location.  Takes the program name as an argument.
 */
static gchar *
fbToolsFindCertXml(
    const gchar  *argv0)
{
    const gchar * const *sysdirs;
    const gchar *fb_pkg_dir;
    gboolean verbose = TRUE;
    gchar  *path;
    gchar  *dir;
    int     i;

    /* cli-provided location */
    if (fbtools_user_opts.cert_element_file_paths) {
        gchar  **filepath = fbtools_user_opts.cert_element_file_paths;
        path = NULL;
        for ( ; *filepath; ++filepath) {
            path = fbToolsCheckXml(*filepath);
            if (path) {
                if (verbose) {
                    g_debug("Checking for %s: Found file '%s'",
                            CERT_IPFIX_BASENAME, path);
                }
                break;
            }
        }
        g_strfreev(fbtools_user_opts.cert_element_file_paths);
        if (path) {
            return path;
        }
    }

    /* the directory ../share/libfixbuf relative to the application's
     * location */
    path = g_path_get_dirname(argv0);
    dir = g_build_filename(path, "..", "share", FBTOOLS_PACKAGE_NAME, NULL);
    g_free(path);
    path = fbToolsCheckXml(dir);
    g_free(dir);
    if (path) {
        if (verbose) {
            g_debug("Checking for %s: Found file '%s'",
                    CERT_IPFIX_BASENAME, path);
        }
        return path;
    }

#if 0
    /* the user's data dir */
    dir = g_build_filename(g_get_user_data_dir(), FBTOOLS_PACKAGE_NAME, NULL);
    path = fbToolsCheckXml(dir);
    g_free(dir);
    if (path) {
        if (verbose) {
            g_debug("Checking for %s: Found file '%s'",
                    CERT_IPFIX_BASENAME, path);
        }
        return path;
    }
#endif  /* 0 */

    /* the compile-time location */
    fb_pkg_dir = FBTOOLS_PACKAGE_DATADIR;
    if (fb_pkg_dir && *fb_pkg_dir) {
        dir = g_build_filename(fb_pkg_dir, NULL);
        path = fbToolsCheckXml(dir);
        g_free(dir);
        if (path) {
            if (verbose) {
                g_debug("Checking for %s: Found file '%s'",
                        CERT_IPFIX_BASENAME, path);
            }
            return path;
        }
    }

    /* system locations */
    path = NULL;
    sysdirs = g_get_system_data_dirs();
    for (i = 0; (sysdirs[i]); ++i) {
        dir = g_build_filename(sysdirs[i], FBTOOLS_PACKAGE_NAME, NULL);
        path = fbToolsCheckXml(dir);
        g_free(dir);
        if (path) {
            if (verbose) {
                g_debug("Checking for %s: Found file '%s'",
                        CERT_IPFIX_BASENAME, path);
            }
            return path;
        }
    }

    fprintf(stderr, "%s: Failed to find '%s'",
            g_get_prgname(), CERT_IPFIX_BASENAME);
    fprintf(stderr, "%s: Use --cert-element-path to specify its location\n",
            g_get_prgname());
    return NULL;
}


/**
 *    If `input_file` is non-NULL, the caller wishes to manage reading the
 *    input itself and no \ref fbCollector_t is associated with the \ref
 *    fBuf_t.
 */
fBuf_t *
fbToolsOpenInput(
    const char    *argv0,
    fbSession_t  **session,
    FILE         **input_file)
{
    FILE          *infile = NULL;
    GError        *err = NULL;
    fbCollector_t *collector = NULL;
    gchar         *cert_xml = NULL;

    model = fbInfoModelAlloc();

    if (!fbtools_user_opts.no_cert_elements) {
        cert_xml = fbToolsFindCertXml(argv0);
        if (NULL == cert_xml) {
            exit(1);
        }

        if (!fbInfoModelReadXMLFile(model, cert_xml, &err)) {
            fprintf(stderr, "%s: Failed to load elements from '%s': %s\n",
                    g_get_prgname(), cert_xml, err->message);
            g_clear_error(&err);
            exit(-1);
        }
        g_free(cert_xml);
    }

    if (fbtools_user_opts.element_files) {
        gchar  **xml_file;
        for (xml_file = fbtools_user_opts.element_files; *xml_file; ++xml_file) {
            if (!fbInfoModelReadXMLFile(model, *xml_file, &err)) {
                fprintf(stderr, "%s: Failed to load elements from '%s': %s\n",
                        g_get_prgname(), *xml_file, err->message);
                g_clear_error(&err);
                exit(-1);
            }
        }
        g_strfreev(fbtools_user_opts.element_files);
    }



    if (fbtools_user_opts.inspec != NULL) {
        if (0 == strcmp("-", fbtools_user_opts.inspec)) {
            infile = stdin;
        } else {
            infile = fopen(fbtools_user_opts.inspec, "r");
            if (infile == NULL) {
                fprintf(stderr, "%s: Failed to open input file '%s': %s\n",
                        g_get_prgname(), fbtools_user_opts.inspec,
                        strerror(errno));
                exit(1);
            }
        }
    } else if (isatty(fileno(stdin))) {
        fprintf(stderr, "%s: No input argument and stdin is a terminal\n",
                g_get_prgname());
        exit(1);
    } else {
        infile = stdin;
    }

    /* Create New Session */
    *session = fbSessionAlloc(model);

    if (input_file) {
        /* Hand infile to the caller; do not create a collector */
        *input_file = infile;
    } else {
        /* Create a Collector */
        collector = fbCollectorAllocFP(NULL, infile);
    }

    /* Allocate Collection Buffer */
    return fBufAllocForCollection(*session, collector);
}


void
fbToolsCloseInput(
    fBuf_t  *fbuf)
{
    if (fbuf) {
        fBufFree(fbuf);
    }

    fbInfoModelFree(model);
    model = NULL;
}



FILE *
fbToolsOpenOutput(
    void)
{
    FILE *out;

    if (NULL == fbtools_user_opts.outspec
        || 0 == strcmp("-", fbtools_user_opts.outspec))
    {
        return stdout;
    }

    out = fopen(fbtools_user_opts.outspec, "w");
    if (NULL == out) {
        fprintf(stderr, "%s: Opening output file %s failed: %s\n",
                g_get_prgname(), fbtools_user_opts.outspec, strerror(errno));
        exit(1);
    }

    return out;
}

gboolean
fbToolsCloseOutput(
    FILE  *fp)
{
    if (fp != stdout && fp != NULL) {
        fclose(fp);
    }
    return TRUE;
}
