/*
 *  Copyright 2014-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *    @file fbtools-lib.h
 *
 *    Helper macros, typedefs, and function definitions; part of libfbtools.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#ifndef _FBTOOLS_LIB_H_
#define _FBTOOLS_LIB_H_
#include "fbtools-autohdr.h"
#include "fbtools-version.h"
#include "fbtools-compat.h"


/**
 *    CERT IPFIX Private Enterprise Number
 */
#define CERT_PEN    6871


/**
 *    String used to wrap long argument descriptions
 */
#define WRAP_OPT "\n                "


/**
 *    How to display string and octetArray values
 */
typedef enum fbToolsFormat_en {
    /* show nothing (only the length) */
    FBTOOLS_FMT_EMPTY,
    /* show as a quoted string with TAB, NEWLINE, CARRIAGE_RETURN,
     * double-quote, and backslash printed as \t, \n, \r, \", \\, and
     * non-ascii encoded as escaped hexadecimal (eg, \\x83) */
    FBTOOLS_FMT_ASCII,
    /* show as a base64-encoded value */
    FBTOOLS_FMT_BASE64,
    /* show as a hexadecimal value with a leading 0x */
    FBTOOLS_FMT_HEXADECIMAL,
    /* show in the style of `hexdump -vC` */
    FBTOOLS_FMT_HEXDUMP,
    /* show as a list of space-separated hexadecimal values */
    FBTOOLS_FMT_OCTETS,
    /* show as a raw unquoted string as in ipfixDump v2, only for FB_STRING */
    FBTOOLS_FMT_RAW,
    /* show as a hexadecimal value if length is <= 8, nothing otherwise; only
     * used for octetArrys */
    FBTOOLS_FMT_SHORTHEX,
    /* alias for FBTOOLS_FMT_STRING, but only valid for --string-format */
    FBTOOLS_FMT_QUOTED,
    /* show as a utf-8 string TAB, NEWLINE, CARRIAGE_RETURN, double-quote, and
     * backslash printed as \t, \n, \r, \", \\, and any non-printable unicode
     * char printed as \uHHHH */
    FBTOOLS_FMT_STRING,
    /* coerce an invalid UTF-8 string to valid Unicode by using the Unicode
     * repplace character U+FFFD for invalid octets, then print as string */
    FBTOOLS_FMT_UNICODE
} fbToolsFormat_t;


/**
 *    Values to pass to fbToolsParseStringOctetFormat() to specify the type of
 *    argument.
 */
typedef enum fbToolsStringOctetType_en {
    /** The argument is always invalid */
    FBTOOLS_ARG_INVALID = 0,
    /** The value was an argument to --octet-format */
    FBTOOLS_ARG_OCTETS = 1,
    /** The value was an argument to --string-format */
    FBTOOLS_ARG_STRING = 2
} fbToolsStringOctetType_t;


/**
 *    Parses `arg_value` for either the --octet-format or --string-format
 *    options.  The caller must pass the type of the argument in `arg_type`
 *    and the name of the option in `option_name` which will be used for error
 *    reporting.
 */
fbToolsFormat_t
fbToolsParseStringOctetFormat(
    fbToolsStringOctetType_t  arg_type,
    const char               *arg_value,
    const char               *option_name);

/**
 *    Adds the --help option to the options group.
 */
void
fbToolsAddOptionsHelp(
    GOptionGroup   *group);

/**
 *    Adds the --version option to the options group.
 */
void
fbToolsAddOptionsVersion(
    GOptionGroup   *group);

/**
 *    Adds the input, output, and element-file options to the options group.
 *
 *    Specifically, adds:
 *
 *    1.  --in and --out for input and output
 *
 *    2.  --cert-element-path to add a search location for cert_ipfix.xml file
 *    and --no-cert-elements to disable loading this file
 *
 *    3.  --element-file to load specific files
 *
 *    4.  --rfc5610 to use elements defined in the input
 */
void
fbToolsAddOptionsIOXml(
    GOptionGroup   *group);

/**
 *    Creates the fbInfoModel_t, the fbSession_t, loads the element files into
 *    the model, and opens the input file.  The session is handed back in the
 *    referent to `session`.
 *
 *    `argv0` is the value of argv[0] to main().  It is used when locating the
 *    CERT element file.
 *
 *    If `input_file` is NULL, the file is used to create an @ref fBuf_t an as
 *    fbCollector_t and the fBuf_t is returned.
 *
 *    If `input_file` is non-NULL, the @ref fBuf_t is not created and the
 *    referent of `input_file` is set to the file.  This may be used when the
 *    caller wants to manage reading of the file.
 */
fBuf_t *
fbToolsOpenInput(
    const char    *argv0,
    fbSession_t  **session,
    FILE         **input_file);

/**
 *    Destroys `fbuf`, the Session, the Model, and closes the input file.
 */
void
fbToolsCloseInput(
    fBuf_t  *fbuf);

/**
 *    Opens and returns the output file.
 */
FILE *
fbToolsOpenOutput(
    void);

/**
 *    Closes the output file.  Returns TRUE on success or FALSE on error.
 */
gboolean
fbToolsCloseOutput(
    FILE  *out);

/**
 *    Returns TRUE if the user provided the --rfc5610 option.
 */
gboolean
fbToolsGetStatusRfc5610(
    void);



/**
 *    Prints an IPv6 address to `fp`.
 */
void
fbToolsPrintIP6Address(
    FILE          *fp,
    const uint8_t *ipaddr);

/**
 *    Prints the fbVarfield_t as ASCII.
 *
 *    Specifically octets in the range 0x20 (space) through 0x7e (tilde) are
 *    printed as ASCII, where 0x22 (quotation mark) and 0x5C (reverse solidus)
 *    are represented as a two-character sequence ("\"" or "\\"), and all
 *    other octets are represented as a five-charcter sequence "\\xhh" where h
 *    is a single hexadecimal digit.
 */
void
fbToolsPrintAscii(
    FILE               *fp,
    const fbVarfield_t *vf);

/**
 *    Prints the fbVarfield_t as UTF-8 encoded Unicode.  The caller must check
 *    the that value is valid UTF-8 before calling this function.
 *
 *    Specifically, writes printable unicode characters to the output and
 *    using a two-character escape sequence for U+0022 (quotation mark, "\""),
 *    U+005C (reverse solidus, "\\"), U+0009 (tab, "\t"), U+000A (line feed,
 *    "\n"), and U+000D (carriage return, "\r").  Non-printable unicode
 *    characters in the range U+0001 to U+FFFF are represented as "\uHHHH" and
 *    those in the range U+10000 to U+10FFFF as the UTF-16 surrogate pair
 *    "\uHHHH\uHHHH".
 */
void
fbToolsPrintUnicode(
    FILE               *fp,
    const fbVarfield_t *vf);



/**
 *    Holds the locations of the template-id and template-name elements within
 *    the template-name template.  The caller is expected to add this to the
 *    template context.
 *
 *    Use fbToolsTemplateNamesExamineTmpl() to create it.
 */
typedef struct fbToolsTemplateName_st fbToolsTemplateName_t;

/**
 *    Initializes the map from template id to template name.
 */
void
fbToolsTemplateNamesInit(
    void);

/**
 *    Destroys the map from template id to template name.
 */
void
fbToolsTemplateNamesDestroy(
    void);

/**
 *    Returns TRUE if any template name records have been seen.
 */
gboolean
fbToolsTemplateNamesSeen(
    void);

/**
 *    Gets the template name given the template ID.  Returns an empty string
 *    if there is no known name for `tid`.
 */
const char *
fbToolsTemplateNamesGetByTid(
    uint16_t    tid);

/**
 *    If `tmpl` describes a template-name record, this function returns an
 *    object to pass to fbToolsTemplateNamesExamineRecord() when the caller
 *    sees a data record that uses `tmpl`.
 *
 *    If `tmpl` is any other type of template, NULL is returned.
 *
 *    It is recommended the caller call this function within the template
 *    callback and store the value on the template context.
 */
fbToolsTemplateName_t *
fbToolsTemplateNamesExamineTmpl(
    const fbTemplate_t   *tmpl);

/**
 *    Extracts the template name from `record`.  The caller may use
 *    fbToolsTemplateNamesGetByTid() to get the name.
 */
void
fbToolsTemplateNamesExamineRecord(
    const fbToolsTemplateName_t *tmpl_metadata,
    const fbRecord_t            *record);


/**
 *    Holds the locations of various info elements within the RFC5610
 *    template.  The caller is expected to add this to the template context.
 *
 *    Use fbToolsRfc5610ExamineTmpl() to create it.
 */
typedef struct fbToolsRfc5610_st fbToolsRfc5610_t;


/**
 *    If `tmpl` describes an info-element (RFC5610) record, this function
 *    returns an object to pass to fbToolsRfc5610ExamineRecord() when the
 *    caller sees a data record that uses `tmpl`.
 *
 *    If `tmpl` is any other type of template, NULL is returned.
 *
 *    It is recommended the caller call this function within the template
 *    callback and store the value on the template context.
 */
fbToolsRfc5610_t *
fbToolsRfc5610ExamineTmpl(
    const fbTemplate_t *tmpl);

/**
 *    If the user provided the --rfc5610 switch to the application, this
 *    function extracts the info-element information from `record` and adds it
 *    to the information model.
 */
gboolean
fbToolsRfc5610ExamineRecord(
    const fbToolsRfc5610_t  *element_metadata,
    const fbRecord_t        *record);


#endif  /* #ifndef _FBTOOLS_LIB_H_ */
