/*
 *  Copyright 2019-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *  @file ipfix2json.c
 *
 *  IPFIX to JSON conversion tool.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#include "ipfix2json.h"

/* CERT IPFIX Private Enterprise Number */
#define CERT_PEN    6871

/* Initial size of the buffer for an in-memory record. */
#define RECBUF_CAPACITY_INITIAL 4096



/*
 *
 *    ------  IDEAS FOR FUTURE WORK  ------
 *
 *
 *    Having some way to format the keys:
 *
 *    %x   TID in hex, lowercase
 *    %X   TID in hex, uppercase
 *    %#x  TID in hex, lowercase, leading 0x
 *    %#X  TID in hex, uppercase, leading 0x
 *
 *    %t   type of the containing BL, STL, or STML
 *    %e   information element of the containing BL, STL, or STML
 *    %n   name of the template (defined via metadata) for this top-level
 *         record, STL, or STML-subrecord; empty string if not defined
 *
 *    %b   IE of the elements within a basicList
 *
 *    %{n{a}{b}}  if name of template is defined, uses string 'a'
 *                else uses string 'b', where a and b may contain formats
 *
 *    %{L{a}{b}}  if IE is defined and not BL, STL, STML, uses string 'a'
 *                else uses string 'b' where a and b may contain formats
 *
 *    %{   -- literal {
 *    %}   -- literal }
 *    %%   -- literal %
 *
 *
 *    --key-string top=
 *    --key-string bl=
 *    --key-string stl=
 *    --key-string stlrec=
 *    --key-string stml=
 *
 *    --key-string bl='%{L{%e}{%b}}' --
 *
 *          uses element name whose type is basic list unless that element
 *          basicList in which case it uses the elements contained in the
 *          basicList
 *
 *    --key-string top=''    --
 *
 *          no keys for top-level record
 *
 *    --key-string top='template%#x(%n)'   --
 *
 *          current
 *
 *    --key-string top='%{n{%n}{template-%x}}'   --
 *
 *          name if available, else the tid
 *
 *
 *    --key-string stl='%{L{%e}{%N{%n}{%#x}}}'   --
 *
 *          name of IE of type STL unless it is subTemplateList, then the name
 *          of template used by the STL if available, else the TID
 *
 *    STL:
 *
 *        foo : [ { bar : cow1, baz : dog1 }, { bar : cow2, baz : dog2 } ]]
 *
 *        foo : { this => { bar : cow1, baz : dog1 },
 *                this => { bar : cow2, baz : dog2 } }
 *
 *
 */


static gchar     *show_argument = NULL;
static gchar     *octet_format_argument = NULL;
static gchar     *string_format_argument = NULL;

static FILE      *outfile = NULL;

static gboolean   print_templates = FALSE;
static gboolean   print_data = TRUE;

GHashTable       *template_names;
GStringChunk     *object_keys;

gboolean          allow_duplicate_keys = FALSE;
fbToolsFormat_t   octet_format = FBTOOLS_FMT_BASE64;
fbToolsFormat_t   string_format = FBTOOLS_FMT_STRING;


/* map from --show argument to variable that should be enabled */
static struct showNameValueMap_st {
    const char  *name;
    gboolean    *value;
} showNameValueMap[] = {
    {"data",            &print_data},
    {"templates",       &print_templates},
    {NULL, NULL}
};


static GOptionEntry id_core_option[] = {
    {"show", 's', 0, G_OPTION_ARG_STRING, &show_argument,
     (WRAP_OPT "Prints only these specified items [data]"
      WRAP_OPT "Choices: data,templates"),
     "csv_list"},
    {"allow-duplicates", '\0', 0, G_OPTION_ARG_NONE, &allow_duplicate_keys,
     (WRAP_OPT "Allows duplicate keys within an object when an element or"
      WRAP_OPT "template appears multiple times"),
     NULL},
    {"octet-format", '\0', 0, G_OPTION_ARG_STRING, &octet_format_argument,
     (WRAP_OPT "Prints octetArray values in this format [base64]"
      WRAP_OPT "Choices: ascii,base64,empty,hexadecimal,octets,none,"
      WRAP_OPT "string,unicode"),
     "format"},
    {"string-format", '\0', 0, G_OPTION_ARG_STRING, &string_format_argument,
     (WRAP_OPT "Prints octetArray values in this format [string]"
      WRAP_OPT "Choices: ascii,base64,empty,hexadecimal,octets,none,"
      WRAP_OPT "string,unicode"),
     "format"},
    { NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
};



/**
 *    Parse the command line options.
 */
static void
idParseOptions(
    int   *argc,
    char **argv[])
{
    GOptionContext *ctx = NULL;
    GOptionGroup   *group = NULL;
    GError         *err = NULL;
    gchar          *app;

    app = g_path_get_basename((*argv)[0]);
    g_set_prgname(app);
    g_free(app);

    ctx = g_option_context_new(" - ipfix2json Options");

    /* Create the main group so we can set user_data to ctx */
    group = g_option_group_new("main", "Options", NULL, ctx, NULL);
    g_option_context_set_main_group(ctx, group);

    /* Create the help option */
    g_option_context_set_help_enabled(ctx, FALSE);
    fbToolsAddOptionsHelp(group);

    /* Add --in, --out, and element options */
    fbToolsAddOptionsIOXml(group);

    /* Add options specific to ipfix2json */
    g_option_group_add_entries(group, id_core_option);

    /* Add --version */
    fbToolsAddOptionsVersion(group);

    if (!g_option_context_parse(ctx, argc, argv, &err)) {
        fprintf(stderr, "%s: Option parsing failed: %s\n",
                g_get_prgname(), err->message);
        g_clear_error(&err);
        exit(1);
    }

    /* check for non-arguments in argv */
    if (1 != *argc) {
        fprintf(stderr, "%s: Unrecognized argument %s\n",
                g_get_prgname(), (*argv)[1]);
        exit(1);
    }

    if (show_argument) {
        gchar      **show_list = g_strsplit(show_argument, ",", -1);
        gchar       *show_item;
        unsigned int i, j;
        print_data = FALSE;
        print_templates = FALSE;

        for (i = 0; show_list[i] != NULL; ++i) {
            show_item = g_strstrip(show_list[i]);
            for (j = 0; showNameValueMap[j].name != NULL; ++j) {
                if (0 == strcmp(show_item, showNameValueMap[j].name)) {
                    *showNameValueMap[j].value = TRUE;
                    break;
                }
            }
            /* error for an invalid argument unless empty */
            if (NULL == showNameValueMap[j].name && '\0' != show_item[0]) {
                fprintf(stderr, "%s: Unrecognized show option: %s\n",
                        g_get_prgname(), show_item);
                g_strfreev(show_list);
                exit(1);
            }
        }
        g_strfreev(show_list);
    }

    if (octet_format_argument) {
        octet_format = fbToolsParseStringOctetFormat(
            FBTOOLS_ARG_OCTETS, octet_format_argument, "octet-format");
        switch (octet_format) {
          case FBTOOLS_FMT_HEXDUMP:
          case FBTOOLS_FMT_SHORTHEX:
            fprintf(stderr, "%s: Unrecognized %s option: %s\n",
                    g_get_prgname(), "octet-format", octet_format_argument);
            exit(1);
          default:
            break;
        }
    }
    if (string_format_argument) {
        string_format = fbToolsParseStringOctetFormat(
            FBTOOLS_ARG_STRING, string_format_argument, "string-format");
        switch (string_format) {
          case FBTOOLS_FMT_HEXDUMP:
          case FBTOOLS_FMT_QUOTED:
          case FBTOOLS_FMT_RAW:
            fprintf(stderr, "%s: Unrecognized %s option: %s\n",
                    g_get_prgname(), "string-format", string_format_argument);
            exit(1);
          default:
            break;
        }
    }

    g_option_context_free(ctx);
}


/**
 *    Free the template context object.  The signature of this
 *    function is given by fbTemplateCtxFree_fn.
 */
static void
idTemplateCtxFree(
    void  *ctx,
    void  *app_ctx)
{
    (void)app_ctx;
    g_free(ctx);
}

/**
 *    Callback invoked when a new template is seen.  The signature of
 *    this function is given by fbNewTemplateCallback_fn.
 */
static void
idTemplateCallback(
    fbSession_t           *session,
    uint16_t               tid,
    fbTemplate_t          *tmpl,
    void                  *app_ctx,
    void                 **ctx,
    fbTemplateCtxFree_fn  *fn)
{
    GError        *err = NULL;
    tmplContext_t *myctx = g_new0(tmplContext_t, 1);

    (void)app_ctx;

    myctx->count = fbTemplateCountElements(tmpl);
    myctx->len = fbTemplateGetIELenOfMemBuffer(tmpl);
    myctx->tid = tid;

    if (print_templates) {
        idPrintTemplate(outfile, tmpl, myctx);
    }

    if (!fbSessionAddTemplate(session, TRUE, tid, tmpl, NULL, &err)) {
        fprintf(stderr, "%s: Error adding template to session: %s\n",
                g_get_prgname(), err->message);
        g_clear_error(&err);
    }

    if (fbTemplateGetOptionsScope(tmpl)) {
        myctx->meta_element = fbToolsRfc5610ExamineTmpl(tmpl);
        myctx->meta_template = fbToolsTemplateNamesExamineTmpl(tmpl);
    }

    *ctx = myctx;
    *fn = idTemplateCtxFree;
}


int
main(
    int    argc,
    char  *argv[])
{
    fBuf_t        *fbuf = NULL;
    GError        *err = NULL;
    gboolean       rc;
    fbSession_t   *session = NULL;
    tmplContext_t *tctx;
    fbRecord_t     record = FB_RECORD_INIT;

    idParseOptions(&argc, &argv);

    if (FALSE == allow_duplicate_keys) {
        /*
         *  To maintain unique JSON keys in each object (record) when an
         *  element is repeated, each object keeps a table of unique keys it
         *  has seen.  The keys are stored as string chunks.
         */
        object_keys = g_string_chunk_new(1024);
    }

    /* Create table to hold template names keyed by TID */
    fbToolsTemplateNamesInit();

    /* Open the input and output files */
    fbuf = fbToolsOpenInput(argv[0], &session, NULL);
    outfile = fbToolsOpenOutput();

    fbSessionAddNewTemplateCallback(session, idTemplateCallback, fbuf);

    /* Allocate buffer for a single in-memory record */
    record.reccapacity = RECBUF_CAPACITY_INITIAL;
    record.rec = g_new(uint8_t, record.reccapacity);

    /* Read until no more data or error */
    for (;;) {
        record.tmpl = fBufNextCollectionTemplate(fbuf, &record.tid, &err);
        if (!record.tmpl) {
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOF)) {
                g_clear_error(&err);
                break;
            }
            fprintf(stderr, "%s: Warning: %s\n",
                    g_get_prgname(), err->message);
            g_clear_error(&err);
            continue;
        }

        tctx = fbTemplateGetContext(record.tmpl);
        record.recsize = tctx->len;

        if (record.recsize > record.reccapacity) {
            do {
                record.reccapacity <<= 1;
            } while (record.recsize > record.reccapacity);
            record.rec = g_renew(uint8_t, record.rec, record.reccapacity);
        }
        memset(record.rec, 0, record.reccapacity);

        rc = fBufNextRecord(fbuf, &record, &err);
        if (FALSE == rc) {
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOF)) {
                fprintf(stderr, "%s: END OF FILE\n", g_get_prgname());
                fBufFree(fbuf);
                g_clear_error(&err);
                break;
            }
            if (!g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOM) &&
                !g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_BUFSZ))
            {
                fprintf(stderr, "%s: Warning: %s\n",
                        g_get_prgname(), err->message);
            }
            g_clear_error(&err);
            continue;
        }

        if (tctx->meta_template) {
            fbToolsTemplateNamesExamineRecord(tctx->meta_template, &record);
        } else if (tctx->meta_element) {
            fbToolsRfc5610ExamineRecord(tctx->meta_element, &record);
        }
        if (!print_data) {
            fBufListFree(record.tmpl, record.rec);
        } else {
            char str_template[TMPL_NAME_BUFSIZ];
            idFormatTemplateId(str_template, record.tid,
                               "__", sizeof(str_template));
            fprintf(outfile, "{%s,", str_template);
            idPrintDataRecord(outfile, &record);
            fprintf(outfile, "}\n");
        }
    }

    g_free(record.rec);

    g_string_chunk_free(object_keys);

    fbToolsCloseInput(fbuf);
    fbToolsCloseOutput(outfile);

    fbToolsTemplateNamesDestroy();

    return 0;
}
