/*
 *  Copyright 2019-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *  @file ipfix2jsonPrint.c
 *
 *  IPFIX to JSON conversion tool.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#include "ipfix2json.h"

#if FIXBUF_CHECK_VERSION(3, 0, 0) && !FIXBUF_CHECK_VERSION(3, 0, 1)
/* work-around broken macro in libfixbuf-3.0.0 */
#undef fbInfoElementGetUnits
#define fbInfoElementGetUnits(ie_)                      \
    (fbInfoElementUnits_t)(FB_IE_UNITS((ie_)->flags))
#endif


static void
idPrintValue(
    FILE                     *fp,
    const fbTemplateField_t  *field,
    fbRecordValue_t          *value,
    GHashTable              **uniq_names);


/**
 *  Looks for 'name' in referent of 'table'.
 *
 *  If found, increments the counter in the table.  If 'namelen' is non-zero,
 *  appends the counter to 'name'.  Returns the updated counter.
 *
 *  If not found, initializes it, adds it to the table, does not modify
 *  'name', and returns 0.
 *
 *  Creates the table if it does not exist.
 */
static unsigned int
idUpdateUniqNameCount(
    GHashTable **table,
    char        *name,
    size_t       namelen)
{
    unsigned int *current;
    gchar        *uniqname;
    size_t        sz;

    if (allow_duplicate_keys) {
        return 0;
    }
    uniqname = g_string_chunk_insert_const(object_keys, name);

    if (*table) {
        /* if current exists in the table, update it in place */
        current = (unsigned int *)g_hash_table_lookup(*table, uniqname);
        if (current) {
            ++*current;
            if (namelen) {
                sz = strlen(name);
                snprintf(name + sz, namelen - sz, "-%u", *current);
            }
            return *current;
        }
    } else {
        *table = g_hash_table_new_full(g_direct_hash, NULL, NULL, &g_free);
    }

    /* add it to the table */
    current = g_new(unsigned int, 1);
    *current = 1;
    g_hash_table_insert(*table, uniqname, current);
    return 0;
}


/**
 *    Puts textual information about the template whose ID is 'tid'
 *    into 'tmpl_str'.  The information is the template ID (both
 *    decimal and hex) and the template name if available.
 *
 *    @param tmpl_str        The buffer to write to
 *    @param tid             The ID of the template to get info of
 *    @param tmpl_str_bufsiz The sizeof the tmpl_str buffer
 *
 */
void
idFormatTemplateId(
    char       *tmpl_str,
    int         tid,
    const char *prefix,
    size_t      tmpl_str_bufsiz)
{
    const char *name = fbToolsTemplateNamesGetByTid(tid);

    if (name[0]) {
        snprintf(tmpl_str, tmpl_str_bufsiz,
                 ("\"%stemplateId\":\"%#06x(%u)\","
                  "\"%stemplateName\":\"%s\""),
                 prefix, tid, tid, prefix, name);
    } else {
        snprintf(tmpl_str, tmpl_str_bufsiz,
                 "\"%stemplateId\":\"%#06x(%u)\"",
                 prefix, tid, tid);
    }
}


/**
 *    Print a textual representation of 'tmpl' to 'fp'.  'ctx' is the
 *    template context created when the template was first read.
 */
void
idPrintTemplate(
    FILE           *fp,
    fbTemplate_t   *tmpl,
    tmplContext_t  *ctx)
{
    const fbTemplateField_t *field = NULL;
    char                     str_template[TMPL_NAME_BUFSIZ];
    fbTemplateIter_t         iter;
    char                     local[20];
    const char              *datatype;
    const char              *units;

    idFormatTemplateId(str_template, ctx->tid, "", sizeof(str_template));
    fprintf(fp, "{%s,\"templateFields\":[", str_template);

    fbTemplateIterInit(&iter, tmpl);
    while ((field = fbTemplateIterNext(&iter))) {

        /* data type value */
        switch (fbTemplateFieldGetType(field)) {
          case 0:
            datatype = "octetArray";
            break;
          case 1:
            datatype = "unsigned8";
            break;
          case 2:
            datatype = "unsigned16";
            break;
          case 3:
            datatype = "unsigned32";
            break;
          case 4:
            datatype = "unsigned64";
            break;
          case 5:
            datatype = "signed8";
            break;
          case 6:
            datatype = "signed16";
            break;
          case 7:
            datatype = "signed32";
            break;
          case 8:
            datatype = "signed64";
            break;
          case 9:
            datatype = "float32";
            break;
          case 10:
            datatype = "float64";
            break;
          case 11:
            datatype = "boolean";
            break;
          case 12:
            datatype = "macAddress";
            break;
          case 13:
            datatype = "string";
            break;
          case 14:
            datatype = "dateTimeSeconds";
            break;
          case 15:
            datatype = "dateTimeMilliseconds";
            break;
          case 16:
            datatype = "dateTimeMicroseconds";
            break;
          case 17:
            datatype = "dateTimeNanoseconds";
            break;
          case 18:
            datatype = "ipv4Address";
            break;
          case 19:
            datatype = "ipv6Address";
            break;
          case 20:
            datatype = "basicList";
            break;
          case 21:
            datatype = "subTemplateList";
            break;
          case 22:
            datatype = "subTemplateMultiList";
            break;
          default:
            snprintf(local, sizeof(local), "Unassigned(%u)",
                     fbTemplateFieldGetType(field));
            datatype = NULL;
            break;
        }

        /* name and ID */
        fprintf(fp,
                ("%s{"
                 "\"informationElementName\":\"%s\","
                 "\"informationElementId\":%d,"),
                ((fbTemplateIterGetPosition(&iter) > 0) ? "," : ""),
                fbTemplateFieldGetName(field),
                fbTemplateFieldGetId(field));

        /* Private enterprise number */
        if (fbTemplateFieldGetPEN(field) != 0) {
            fprintf(fp,
                    "\"privateEnterpriseNumber\":%u,",
                    fbTemplateFieldGetPEN(field));
        }
        /* Repeat or Multi-Use Index */
        if (fbTemplateFieldGetRepeat(field) != 0) {
            fprintf(fp,
                    "\"informationElementIndex\":%u,",
                    fbTemplateFieldGetRepeat(field));
        }

        /* Length, Data-type */
        fprintf(fp,
                ("\"templateFieldLength\":%u,"
                 "\"informationElementDataType\":\"%s\""),
                fbTemplateFieldGetLen(field), datatype);

        /* units */
        switch (fbInfoElementGetUnits(fbTemplateFieldGetIE(field))) {
          case 0:
            units = NULL;
            break;
          case 1:
            units = "bits";
            break;
          case 2:
            units = "octets";
            break;
          case 3:
            units = "packets";
            break;
          case 4:
            units = "flows";
            break;
          case 5:
            units = "seconds";
            break;
          case 6:
            units = "milliseconds";
            break;
          case 7:
            units = "microseconds";
            break;
          case 8:
            units = "nanoseconds";
            break;
          case 9:
            units = "4-octet words";
            break;
          case 10:
            units = "messages";
            break;
          case 11:
            units = "hops";
            break;
          case 12:
            units = "entries";
            break;
          case 13:
            units = "frames";
            break;
          case 14:
            units = "ports";
            break;
          case 15:
            units = "inferred";
            break;
          default:
            snprintf(local, sizeof(local), "Unassigned(%u)",
                     fbInfoElementGetUnits(fbTemplateFieldGetIE(field)));
            units = local;
            break;
        }
        if (units) {
            fprintf(fp, ",\"informationElementUnits\":\"%s\"", units);
        }

        fprintf(fp, "}");
    }
    fprintf(fp, "]}\n");
}


/**
 *    Print a textual representation of 'entry' to 'fp'.
 */
static void
idPrintSTMLEntry(
    FILE                           *fp,
    fbSubTemplateMultiListEntry_t  *entry)
{
    gboolean   first = TRUE;
    char       str_template[TMPL_NAME_BUFSIZ];
    fbRecord_t subrec = FB_RECORD_INIT;

    subrec.tid = fbSubTemplateMultiListEntryGetTemplateID(entry);
    subrec.tmpl = fbSubTemplateMultiListEntryGetTemplate(entry);
    idFormatTemplateId(str_template, subrec.tid, "__", sizeof(str_template));

    while ((subrec.rec = fbSTMLEntryNext(uint8_t, entry, subrec.rec))) {
        if (!first) {
            fprintf(fp, ",");
        } else {
            first = FALSE;
        }
        fprintf(fp, "{%s,", str_template);
        idPrintDataRecord(fp, &subrec);
        fprintf(fp, "}");
    }
}


/**
 *    Print a textual representation of 'stl' to 'fp'.
 */
static void
idPrintSTL(
    FILE                       *fp,
    const fbSubTemplateList_t  *stl)
{
    char       str_template[TMPL_NAME_BUFSIZ];
    gboolean   first = TRUE;
    fbRecord_t subrec = FB_RECORD_INIT;

    subrec.tid = fbSubTemplateListGetTemplateID(stl);
    subrec.tmpl = fbSubTemplateListGetTemplate(stl);
    idFormatTemplateId(str_template, subrec.tid, "__", sizeof(str_template));

    fprintf(fp, "[");
    while ((subrec.rec = fbSTLNext(uint8_t, stl, subrec.rec))) {
        if (!first) {
            fprintf(fp, ",");
        } else {
            first = FALSE;
        }
        fprintf(fp, "{%s,", str_template);
        idPrintDataRecord(fp, &subrec);
        fprintf(fp, "}");
    }
    fprintf(fp, "]");;
}


/**
 *    Print a textual representation of 'stml' to 'fp'.
 */
static void
idPrintSTML(
    FILE                            *fp,
    const fbSubTemplateMultiList_t  *stml)
{
    fbSubTemplateMultiListEntry_t *entry = NULL;
    gboolean    first = TRUE;

    fprintf(fp, "[");
    while ((entry = fbSTMLNext(stml, entry))) {
        if (!first) {
            fprintf(fp, ",");
        } else {
            first = FALSE;
        }
        idPrintSTMLEntry(fp, entry);
    }

    fprintf(fp, "]");
}


/**
 *    Print a textual representation of 'bl' to 'fp'.
 */
static void
idPrintBL(
    FILE                     *fp,
    const fbBasicList_t      *bl,
    GHashTable              **uniq_names)
{
    const fbTemplateField_t *field;
    fbRecordValue_t          value = FB_RECORD_VALUE_INIT;
    GHashTable              *uniq_children = NULL;
    unsigned int             i;
    unsigned int             count;

    field = fbBasicListGetTemplateField(bl);
    count = idUpdateUniqNameCount(
        uniq_names, (char *)fbTemplateFieldGetName(field), 0);
    if (0 == count) {
        fprintf(fp, "\"%s\":[", fbTemplateFieldGetName(field));
    } else {
        fprintf(fp, "\"%s-%u\":[", fbTemplateFieldGetName(field), count);
    }

    for (i = 0; fbBasicListGetIndexedRecordValue(bl, i, &value); ++i) {
        if (i > 0) {
            fprintf(fp, ",");
        }
        idPrintValue(fp, field, &value, &uniq_children);
    }
    if (uniq_children) {
        g_hash_table_destroy(uniq_children);
    }

    fprintf(fp, "]");
}


/**
 *    Prints `value` to `fp` using format `format` where `value` is of type
 *    string or octet array.  Clears `value`.
 */
static void
idPrintStringOrOctetArray(
    FILE               *fp,
    fbRecordValue_t    *value,
    fbToolsFormat_t  format)
{
    size_t i;

    switch (format) {
      case FBTOOLS_FMT_EMPTY:
        fprintf(fp, "\"\"");
        break;

      case FBTOOLS_FMT_ASCII:
        fbToolsPrintAscii(fp, &value->v.varfield);
        break;

      case FBTOOLS_FMT_BASE64:
        {
            char *b64buf = g_base64_encode(value->v.varfield.buf,
                                           value->v.varfield.len);
            fprintf(fp, "\"%s\"", b64buf);
            g_free(b64buf);
        }
        break;

      case FBTOOLS_FMT_HEXADECIMAL:
        fprintf(fp, "\"");
        for (i = 0; i < value->v.varfield.len; ++i) {
            fprintf(fp, "%02x", value->v.varfield.buf[i]);
        }
        fprintf(fp, "\"");
        break;

      case FBTOOLS_FMT_OCTETS:
        if (0 == value->v.varfield.len) {
            fprintf(fp, "\"\"");
        } else {
            fprintf(fp, "\"%02x", value->v.varfield.buf[0]);
            for (i = 1; i < value->v.varfield.len; ++i) {
                fprintf(fp, " %02x", value->v.varfield.buf[i]);
            }
            fprintf(fp, "\"");
        }
        break;

      case FBTOOLS_FMT_STRING:
        if (g_utf8_validate((gchar *)value->v.varfield.buf,
                            value->v.varfield.len, NULL))
        {
            fbToolsPrintUnicode(fp, &value->v.varfield);
        } else {
            fbToolsPrintAscii(fp, &value->v.varfield);
        }
        break;

      case FBTOOLS_FMT_UNICODE:
        if (g_utf8_validate((gchar *)value->v.varfield.buf,
                            value->v.varfield.len, NULL))
        {
            fbToolsPrintUnicode(fp, &value->v.varfield);
        } else {
            /* Create a validated form of the string using the unicode
             * replacement character U+FFFD for the invalid parts. */
            fbVarfield_t vf;
            vf.buf = (uint8_t *)g_utf8_make_valid(
                (gchar *)value->v.varfield.buf, value->v.varfield.len);
            vf.len = strlen((const char *)vf.buf);
            fbToolsPrintUnicode(fp, &vf);
            g_free(vf.buf);
        }
        break;

      case FBTOOLS_FMT_HEXDUMP:
      case FBTOOLS_FMT_QUOTED:
      case FBTOOLS_FMT_RAW:
      case FBTOOLS_FMT_SHORTHEX:
        g_error("Unexpected value %u used in switch", format);
    }

    fbRecordValueClear(value);
}



/**
 *    Print the value of element 'field' to 'fp'.  The value is given in
 *    'val'.
 */
static void
idPrintValue(
    FILE                     *fp,
    const fbTemplateField_t  *field,
    fbRecordValue_t          *value,
    GHashTable              **uniq_names)
{
    struct tm time_tm;
    char      timebuf[32];

    switch (fbTemplateFieldGetType(field)) {
      case FB_BOOL:
      case FB_UINT_8:
      case FB_UINT_16:
      case FB_UINT_32:
      case FB_UINT_64:
        fprintf(fp, "%" PRIu64, value->v.u64);
        break;

      case FB_INT_8:
      case FB_INT_16:
      case FB_INT_32:
      case FB_INT_64:
        fprintf(fp, "%" PRId64, value->v.s64);
        break;

      case FB_IP4_ADDR:
        fprintf(fp, "\"%u.%u.%u.%u\"",
                (value->v.ip4 >> 24), (value->v.ip4 >> 16) & 0xFF,
                (value->v.ip4 >> 8) & 0xFF, value->v.ip4 & 0xFF);
        break;
      case FB_IP6_ADDR:
        fprintf(fp, "\"");
        fbToolsPrintIP6Address(fp, value->v.ip6);
        fprintf(fp, "\"");
        break;
      case FB_FLOAT_64:
      case FB_FLOAT_32:
        fprintf(fp, "%.8g", value->v.dbl);
        break;

      case FB_DT_SEC:
        /* "%Y-%m-%d %H:%M:%S" */
        strftime(timebuf, sizeof(timebuf), "%F %T",
                 gmtime_r(&value->v.dt.tv_sec, &time_tm));
        fprintf(fp, "\"%sZ\"", timebuf);
        break;
      case FB_DT_MILSEC:
        strftime(timebuf, sizeof(timebuf), "%F %T",
                 gmtime_r(&value->v.dt.tv_sec, &time_tm));
        fprintf(fp, "\"%s.%03ldZ\"", timebuf, value->v.dt.tv_nsec / 1000000);
        break;
      case FB_DT_MICROSEC:
        strftime(timebuf, sizeof(timebuf), "%F %T",
                 gmtime_r(&value->v.dt.tv_sec, &time_tm));
        fprintf(fp, "\"%s.%06ldZ\"", timebuf, value->v.dt.tv_nsec / 1000);
        break;
      case FB_DT_NANOSEC:
        strftime(timebuf, sizeof(timebuf), "%F %T",
                 gmtime_r(&value->v.dt.tv_sec, &time_tm));
        fprintf(fp, "\"%s.%09ldZ\"", timebuf, value->v.dt.tv_nsec);
        break;

      case FB_BASIC_LIST:
        idPrintBL(fp, value->v.bl, uniq_names);
        fbBasicListClear((fbBasicList_t *)value->v.bl);
        break;
      case FB_SUB_TMPL_LIST:
        idPrintSTL(fp, value->v.stl);
        fbSubTemplateListClear((fbSubTemplateList_t *)value->v.stl);
        break;
      case FB_SUB_TMPL_MULTI_LIST:
        idPrintSTML(fp, value->v.stml);
        fbSubTemplateMultiListClear((fbSubTemplateMultiList_t *)value->v.stml);
        break;

      case FB_MAC_ADDR:
        fprintf(fp, "\"%02x:%02x:%02x:%02x:%02x:%02x\"",
                value->v.mac[0], value->v.mac[1], value->v.mac[2],
                value->v.mac[3], value->v.mac[4], value->v.mac[5]);
        break;

      case FB_STRING:
        idPrintStringOrOctetArray(fp, value, string_format);
        break;

      case FB_OCTET_ARRAY:
        idPrintStringOrOctetArray(fp, value, octet_format);
        break;
    }
}


/**
 *    Print a textual representation of a record to 'fp'.  The
 *    record's template is 'tmpl', its data is given by 'buffer'.
 */
void
idPrintDataRecord(
    FILE        *fp,
    fbRecord_t  *record)
{
    const fbTemplateField_t *field = NULL;
    fbRecordValue_t          value = FB_RECORD_VALUE_INIT;
    fbTemplateIter_t         iter;
    gboolean                 first = TRUE;
    GHashTable              *uniq_names = NULL;
    unsigned int             count = 0;

    fbTemplateIterInit(&iter, record->tmpl);
    while ((field = fbTemplateIterNext(&iter))) {
        if (fbTemplateFieldCheckIdent(field, 0, 210)) {
            /* paddingOctets */
            continue;
        }
        if (!first) {
            fprintf(fp, ",");
        } else {
            first = FALSE;
        }
        if (FB_BASIC_LIST != fbTemplateFieldGetType(field)) {
            count = idUpdateUniqNameCount(
                &uniq_names, (char *)fbTemplateFieldGetName(field), 0);
            if (0 == count) {
                fprintf(fp, "\"%s\":", fbTemplateFieldGetName(field));
            } else {
                fprintf(fp, "\"%s-%u\":",
                        fbTemplateFieldGetName(field), count);
            }
        }
        fbRecordGetValueForField(record, field, &value);
        idPrintValue(fp, field, &value, &uniq_names);
        /* fflush(fp); */
    }

    if (uniq_names) {
        g_hash_table_destroy(uniq_names);
    }
}
