/*
 *  Copyright 2014-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/**
 *  @file ipfixDump.c
 *
 *  Prints an ipfix file in a human-readable form to outspec or stdout.
 */
/*
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  fixbuf-tools 4.0
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
 *  IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
 *  FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
 *  OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
 *  MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
 *  TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1024
 *  @DISTRIBUTION_STATEMENT_END@
 */

#include "ipfixDump.h"

/* Initial size of the buffer for an in-memory record. */
#define RECBUF_CAPACITY_INITIAL 4096

/* Flags that indicate the types of Sets in a Message  */
#define TMPL_MSG 0x1
#define DATA_MSG 0x2


/* Check whether TemplateField is a particular InfoElement */
#define FIELD_IS_flowEndMilliseconds(f) \
    fbTemplateFieldCheckIdent((f), 0, 153)
#define FIELD_IS_flowEndMicroseconds(f) \
    fbTemplateFieldCheckIdent((f), 0, 155)
#define FIELD_IS_flowEndNanoseconds(f) \
    fbTemplateFieldCheckIdent((f), 0, 157)


static gboolean     yaf = FALSE;
static FILE        *outfile = NULL;
static gboolean     opt_templates = FALSE;
static gboolean     opt_data = FALSE;
static gboolean     opt_file_times = FALSE;
static gboolean     opt_template_counts = FALSE;
static gchar       *show_argument = NULL;
static gchar       *octet_format_argument = NULL;
static gchar       *string_format_argument = NULL;
static gchar       *suppress_argument = NULL;

static gboolean     print_templates = TRUE;
static gboolean     print_data = TRUE;
static gboolean     print_messages = FALSE;
static gboolean     print_file_times = FALSE;
static gboolean     print_template_counts = FALSE;
static gboolean     print_unused_counts = FALSE;

static unsigned int msg_count = 0;
static unsigned int msg_rec_count = 0;
static unsigned int msg_tmpl_count = 0;
static unsigned int tmpl_count = 0;
static gboolean     eom = TRUE;

gboolean            suppress_record_number = FALSE;
static gboolean     suppress_export_time = FALSE;
static gboolean     suppress_message_length = FALSE;
static gboolean     suppress_sequence_number = FALSE;
static gboolean     suppress_stream_offset = FALSE;
static gboolean     suppress_messages = FALSE;

fbToolsFormat_t     octet_array_format = FBTOOLS_FMT_SHORTHEX;
fbToolsFormat_t     string_format = FBTOOLS_FMT_QUOTED;

/* Holds tmplContext_t pointers; the template-count option uses this to
 * process all templates */
static GPtrArray   *tmpl_ctx_array;

static uint32_t     sequence_number = 0;
static size_t       msglen;
static size_t       offset = 0;

/* Whether the current IPFIX message holds template record sets (TMPL_MSG)
 * and/or data record sets (DATA_MSG) */
static int          curr_msg_set_types;

/* used to keep track of message and record timestamps when --file-times is
 * enabled */
typedef struct fileTimes_st {
    uint64_t       first_rectime;
    uint64_t       final_rectime;
    uint64_t       max_rectime;
    uint64_t       min_rectime;
    uint32_t       start_msg_secs;
    uint32_t       last_msg_secs;
} fileTimes_t;

/* map from --show argument to variable that should be enabled */
static struct showNameValueMap_st {
    const char  *name;
    gboolean    *value;
} showNameValueMap[] = {
    {"data",            &print_data},
    {"file-times",      &print_file_times},
    {"messages",        &print_messages},
    {"templates",       &print_templates},
    {"template-counts", &print_template_counts},
    {"unused-counts",   &print_unused_counts},
    {NULL, NULL}
};

/* map from --supress argument to variable that should be enabled */
static struct suppressNameValueMap_st {
    const char  *name;
    gboolean    *value;
} suppressNameValueMap[] = {
    {"export-time",     &suppress_export_time},
    {"message-length",  &suppress_message_length},
    {"messages",        &suppress_messages},
    {"record-number",   &suppress_record_number},
    {"sequence-number", &suppress_sequence_number},
    {"stream-offset",   &suppress_stream_offset},
    {NULL, NULL}
};



static GOptionEntry id_core_option[] = {
    {"yaf", '\0', 0, G_OPTION_ARG_NONE, &yaf,
     (WRAP_OPT "Does nothing; provided for backward compatibility"),
     NULL},
    {"show", 's', 0, G_OPTION_ARG_STRING, &show_argument,
     (WRAP_OPT "Prints only these specified items [data,templates]"
      WRAP_OPT "Choices: data,templates,messages,file-times,"
      WRAP_OPT "template-counts,unused-counts"),
     "csv_list"},
    {"templates", 't', 0, G_OPTION_ARG_NONE, &opt_templates,
     (WRAP_OPT "Prints only IPFIX template records that are present and the"
      WRAP_OPT "message header and footer for messages that contain them"),
     NULL},
    {"data", 'd', 0, G_OPTION_ARG_NONE, &opt_data,
     (WRAP_OPT "Prints only IPFIX data records that are present and the message"
      WRAP_OPT "header and footer for messages that contain them"),
     NULL},
    {"file-times", 'f', 0, G_OPTION_ARG_NONE, &opt_file_times,
     (WRAP_OPT "Prints only file timestamps giving the export time of the first"
      WRAP_OPT "and last IPFIX message, and the record timestamps of the first,"
      WRAP_OPT "last, oldest, and youngest record"),
     NULL},
    {"template-counts", 'c', 0, G_OPTION_ARG_NONE, &opt_template_counts,
     (WRAP_OPT "Prints only template statistics giving the count and sizes of"
      WRAP_OPT "templates that were used by data records (count > 0"),
     NULL},
    {"octet-format", '\0', 0, G_OPTION_ARG_STRING, &octet_format_argument,
     (WRAP_OPT "Prints octetArray values in this format [short-hex]"
      WRAP_OPT "Choices: ascii,base64,empty,hexadecimal,hexdump,octets,none,"
      WRAP_OPT "short-hex,string,unicode"),
     "format"},
    {"string-format", '\0', 0, G_OPTION_ARG_STRING, &string_format_argument,
     (WRAP_OPT "Prints string values in this format [string]"
      WRAP_OPT "Choices: ascii,base64,empty,hexadecimal,hexdump,octets,none,"
      WRAP_OPT "quoted,raw,string,unicode"),
     "format"},
    {"suppress", '\0', 0, G_OPTION_ARG_STRING, &suppress_argument,
     (WRAP_OPT "Disables display of these aspects of the stream"
      WRAP_OPT "Choices: export-time,message-length,record-number,"
      WRAP_OPT "sequence-number,stream-offset,messages"),
     "csv_list"},
    { NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
};


/**
 *    Parse the command line options.
 */
static void
idParseOptions(
    int   *argc,
    char **argv[])
{
    GOptionContext *ctx = NULL;
    GOptionGroup   *group = NULL;
    GError         *err = NULL;
    gchar          *app;
    int             num_only_args = 0;

    app = g_path_get_basename((*argv)[0]);
    g_set_prgname(app);
    g_free(app);

    ctx = g_option_context_new(" - ipfixDump Options");

    /* Create the main group so we can set user_data to ctx */
    group = g_option_group_new("main", "Options", NULL, ctx, NULL);
    g_option_context_set_main_group(ctx, group);

    /* Create the help option */
    g_option_context_set_help_enabled(ctx, FALSE);
    fbToolsAddOptionsHelp(group);

    /* Add --in, --out, and element options */
    fbToolsAddOptionsIOXml(group);

    /* Add options specific to ipfix2json */
    g_option_group_add_entries(group, id_core_option);

    /* Add --version */
    fbToolsAddOptionsVersion(group);

    if (!g_option_context_parse(ctx, argc, argv, &err)) {
        fprintf(stderr, "%s: Option parsing failed: %s\n",
                g_get_prgname(), err->message);
        g_clear_error(&err);
        exit(1);
    }

    if (opt_template_counts) {
        print_data = FALSE;
        print_templates = FALSE;
        print_template_counts = TRUE;
        ++num_only_args;
    }

    if (opt_file_times) {
        print_data = FALSE;
        print_templates = FALSE;
        print_file_times = TRUE;
        ++num_only_args;
    }

    if (opt_data) {
        print_templates = FALSE;
        ++num_only_args;
    }

    if (opt_templates) {
        print_data = FALSE;
        ++num_only_args;
    }

    if (num_only_args > 1) {
        fprintf(stderr, "%s: Using multiple specific print arguments is not "
                "allowed, use --show instead.\n",
                g_get_prgname());
        exit(1);
    }

    /* parse the comma separated value argument to --show */
    if (show_argument != NULL) {
        gchar      **show_list = g_strsplit(show_argument, ",", -1);
        gchar       *show_item;
        unsigned int i, j;
        print_data = FALSE;
        print_templates = FALSE;
        print_file_times = FALSE;
        print_template_counts = FALSE;

        for (i = 0; show_list[i] != NULL; ++i) {
            show_item = g_strstrip(show_list[i]);
            for (j = 0; showNameValueMap[j].name != NULL; ++j) {
                if (0 == strcmp(show_item, showNameValueMap[j].name)) {
                    *showNameValueMap[j].value = TRUE;
                    break;
                }
            }
            /* error for an invalid argument unless empty */
            if (NULL == showNameValueMap[j].name && '\0' != show_item[0]) {
                fprintf(stderr, "%s: Unrecognized show option: %s\n",
                        g_get_prgname(), show_item);
                g_strfreev(show_list);
                exit(1);
            }
        }
        g_strfreev(show_list);

        if (print_unused_counts) {
            print_template_counts = TRUE;
        }
    }

    /* parse the single arguments to --octet-format and --string-format */
    if (octet_format_argument) {
        octet_array_format = fbToolsParseStringOctetFormat(
            FBTOOLS_ARG_OCTETS, octet_format_argument, "octet-format");
    }
    if (string_format_argument) {
        string_format = fbToolsParseStringOctetFormat(
            FBTOOLS_ARG_STRING, string_format_argument, "string-format");
    }

    /* parse the comma separated value argument to --suppress */
    if (suppress_argument) {
        gchar      **suppress_list = g_strsplit(suppress_argument, ",", -1);
        gchar       *item;
        unsigned int i, j;
        for (i = 0; suppress_list[i] != NULL; ++i) {
            item = g_strstrip(suppress_list[i]);
            for (j = 0; suppressNameValueMap[j].name != NULL; ++j) {
                if (0 == strcmp(item, suppressNameValueMap[j].name)) {
                    *suppressNameValueMap[j].value = TRUE;
                    break;
                }
            }
            /* warn about invalid argument unless empty */
            if (NULL == suppressNameValueMap[j].name && '\0' != item[0]) {
                fprintf(stderr, "%s: Unrecognized suppress option: %s\n",
                        g_get_prgname(), item);
                g_strfreev(suppress_list);
                exit(1);
            }
        }
        g_strfreev(suppress_list);
    }

    /* check for non-arguments in argv */
    if (1 != *argc) {
        fprintf(stderr, "%s: Unrecognized argument %s\n",
                g_get_prgname(), (*argv)[1]);
        exit(1);
    }

    g_option_context_free(ctx);
}


/**
 *    Print statistics for the current message if the msg_count is
 *    non-zero.
 */
static void
idCloseMessage(
    void)
{
    const char *prefix = "\n*** Msg Stats:";

    /* the 'msg_count' check is to avoid creating output before the first
     * message is seen. */
    if (msg_count && !suppress_messages) {
        if (print_messages || (msg_rec_count && print_data)) {
            fprintf(outfile, "%s %u Data Records", prefix, msg_rec_count);
            prefix = ",";
        }
        if (print_messages || (msg_tmpl_count && print_templates)) {
            fprintf(outfile, "%s %u Template Records", prefix, msg_tmpl_count);
            prefix = ",";
        }
        if ('\n' != prefix[0]) {
            fprintf(outfile, " ***\n\n");
        }
    }
    tmpl_count += msg_tmpl_count;
}


/**
 *    Print a textual representation of an IPFIX message header to
 *    `outfile`.
 */
static void
idPrintHeader(
    fBuf_t          *fbuf,
    const uint32_t   seq_number,
    const size_t     msg_length,
    const size_t     stream_offset)
{
    const fbSession_t *session;
    long epochtime;
    struct tm          time_tm;
    char timebuf[32];

    session = fBufGetSession(fbuf);

    if (suppress_export_time) {
        snprintf(timebuf, sizeof(timebuf), "%19s", "");
    } else {
        /* "%Y-%m-%d %H:%M:%S" */
        epochtime = fBufGetExportTime(fbuf);
        strftime(timebuf, sizeof(timebuf), "%F %T",
                 gmtime_r(&epochtime, &time_tm));
    }

    fprintf(outfile, "--- Message Header ---\n");
    fprintf(outfile, "export time: %s\t", timebuf);
    fprintf(outfile, "observation domain id: %u\n",
            fbSessionGetDomain(session));
    if (!suppress_message_length) {
        fprintf(outfile, "message length: %-16zu\t", msg_length);
    } else {
        fprintf(outfile, "message length: %16s\t", "");
    }
    if (!suppress_sequence_number) {
        fprintf(outfile, "sequence number: %u (%#x)\n",
                seq_number, seq_number);
    } else {
        fprintf(outfile, "sequence number:\n");
    }
    if (!suppress_stream_offset) {
        fprintf(outfile, "stream offset: %zu\n", stream_offset);
    }
}


/**
 *    Close the current message, print the header for the new message,
 *    and reset the message counters.
 */
static void
idNewMessage(
    fBuf_t  *fbuf)
{
    idCloseMessage();

    if (!suppress_messages &&
        (print_messages ||
         (print_data && (curr_msg_set_types & DATA_MSG)) ||
         (print_templates && (curr_msg_set_types & TMPL_MSG))))
    {
        idPrintHeader(fbuf, sequence_number, msglen, offset - msglen);
    }

    eom = FALSE;
    /* reset msg counters */
    msg_rec_count = 0;
    msg_tmpl_count = 0;
    ++msg_count;
}


/**
 *    Callback invoked when a new template is seen.  The signature of
 *    this function is given by fbNewTemplateCallback_fn.
 */
static void
idTemplateCallback(
    fbSession_t           *session,
    uint16_t               tid,
    fbTemplate_t          *tmpl,
    void                  *app_ctx,
    void                 **ctx,
    fbTemplateCtxFree_fn  *fn)
{
    const fbTemplateField_t *field;
    GError           *err = NULL;
    tmplContext_t    *myctx;
    fbTemplateIter_t  iter;
    gboolean          need_time = FALSE;
    int               len;

    myctx = g_new0(tmplContext_t, 1);
    g_ptr_array_add(tmpl_ctx_array, myctx);

    if (eom) {
        idNewMessage((fBuf_t *)app_ctx);
    }

    myctx->elem_count = fbTemplateCountElements(tmpl);
    myctx->scope = fbTemplateGetOptionsScope(tmpl);
    myctx->len = fbTemplateGetIELenOfMemBuffer(tmpl);
    myctx->tid = tid;
    myctx->iename_width = MIN_IENAME_WIDTH;
    myctx->min_octets = UINT16_MAX;
    /* start as fixed; set to false when a varlen element is found */
    myctx->is_fixedlen = TRUE;

    if (!fbSessionAddTemplate(session, TRUE, tid, tmpl, NULL, &err)) {
        fprintf(stderr, "%s: Error adding template to session: %s\n",
                g_get_prgname(), err->message);
        g_clear_error(&err);
    }

    /* which element(s) to look for in the template */
    if (0 == myctx->scope) {
        /* Look for end-time in all non-Options records */
        need_time = TRUE;
    } else {
        myctx->meta_element = fbToolsRfc5610ExamineTmpl(tmpl);
        myctx->meta_template = fbToolsTemplateNamesExamineTmpl(tmpl);
    }

    /* examine template, looking for specific fields and also finding the
     * length of the longest IE name */
    fbTemplateIterInit(&iter, tmpl);
    while ((field = fbTemplateIterNext(&iter))) {
        len = strlen(fbTemplateFieldGetName(field));
        if (len > myctx->iename_width) {
            myctx->iename_width = len;
        }
        if (fbTemplateFieldGetLen(field) != FB_IE_VARLEN) {
            myctx->fixed_octets += fbTemplateFieldGetLen(field);
        } else {
            myctx->is_fixedlen = FALSE;
            myctx->fixed_octets += 3;
        }
        if (need_time
            && (FIELD_IS_flowEndMilliseconds(field)
                || FIELD_IS_flowEndMicroseconds(field)
                || FIELD_IS_flowEndNanoseconds(field)))
        {
            myctx->end_time_index = *field;
            myctx->contains_end_time = TRUE;
            need_time = FALSE;
        }
    }

    myctx->iename_width *= -1;
    if (print_templates) {
        idPrintTemplate(outfile, tmpl, myctx);
    }

    ++msg_tmpl_count;
    *ctx = myctx;
    *fn = NULL;
}


/**
 *  Updates the record-length statistics on `tctx` with `reclen`.
 */
void
tmplContextUpdateReclenStats(
    tmplContext_t  *tctx,
    uint32_t        reclen)
{
    tctx->all_octets += reclen;
    if (reclen < tctx->min_octets) {
        tctx->min_octets = reclen;
    }
    if (reclen > tctx->max_octets) {
        tctx->max_octets = reclen;
    }
}


/**
 *  Sorts the tmplContext_t* elements in the tmpl_ctx_array by template-id
 *  prior to printing the template counts.
 */
static int
idTemplateContextSort(
    gconstpointer  v_a,
    gconstpointer  v_b)
{
    const tmplContext_t *a = *(const tmplContext_t **)v_a;
    const tmplContext_t *b = *(const tmplContext_t **)v_b;

    return ((a->tid < b->tid)
            ? -1
            : (a->tid - b->tid));
}


/**
 *  Recursively visits all subrecords to calculate the IPFIX-length of
 *  `record` (including its subrecords) and to count the number of times each
 *  template is used by this record and its subrecords.  Updates the
 *  min-record, max-record, and all-rec-length values on `tctx` and returns
 *  the octet length of `record`.
 *
 *  This function is called only when print_data is FALSE and
 *  print_template_counts is TRUE.
 */
static uint32_t
idUpdateTemplateCounts(
    tmplContext_t     *tctx,
    const fbRecord_t  *record)
{
    const fbTemplateField_t *field;
    fbRecord_t        subrec = FB_RECORD_INIT;
    fbTemplateIter_t  iter;
    tmplContext_t    *subtctx;
    uint32_t          reclen;

    ++tctx->rec_count;
    if (tctx->is_fixedlen) {
        return tctx->fixed_octets;
    }

    reclen = tctx->fixed_octets;

    /* Handle any varlen fields and visit the ST(M)L sub-records.
     *
     * Reminder: the 3 bytes that hold the length of the varfield are
     * already accounted for in the fixed_octets */

    fbTemplateIterInit(&iter, record->tmpl);
    while ((field = fbTemplateIterNext(&iter))) {
        switch (fbTemplateFieldGetType(field)) {
          case FB_STRING:
          case FB_OCTET_ARRAY:
            if (FB_IE_VARLEN == fbTemplateFieldGetLen(field)) {
                fbVarfield_t *vf = (fbVarfield_t *)(
                    record->rec + fbTemplateFieldGetOffset(field));
                reclen += vf->len;
            }
            break;
          case FB_BASIC_LIST:
            {
                fbBasicList_t *bl;
                const fbInfoElement_t *ie;
                uint16_t len;

                bl = ((fbBasicList_t *)
                      (record->rec + fbTemplateFieldGetOffset(field)));
                ie = fbBasicListGetInfoElement(bl);
                len = fbBasicListGetElementLength(bl);

                /* BL includes semantic(1), elem-id(2), elem-length(2), and
                 * enterprise-id(4) for non-standard elements */
                reclen += 5 + 4 * (0 != fbInfoElementGetPEN(ie));
                if (FB_IE_VARLEN != len) {
                    reclen += len * fbBasicListCountElements(bl);
                } else if (FB_STRING == fbInfoElementGetType(ie) ||
                           FB_OCTET_ARRAY == fbInfoElementGetType(ie))
                {
                    fbVarfield_t *vf = NULL;
                    while ((vf = fbBLNext(fbVarfield_t, bl, vf)) != NULL) {
                        reclen += 3 + vf->len;
                    }
                } else {
                    /* yaf doesn't support lists in a basicList */
                }
                fbBasicListClear(bl);
            }
            break;
          case FB_SUB_TMPL_LIST:
            {
                fbSubTemplateList_t *stl;

                stl = ((fbSubTemplateList_t *)
                       (record->rec + fbTemplateFieldGetOffset(field)));

                /* STL includes semantic(1), tmpl-id(2) */
                reclen += 3;
                subrec.tid = fbSubTemplateListGetTemplateID(stl);
                subrec.tmpl = fbSubTemplateListGetTemplate(stl);
                subrec.rec = NULL;
                subtctx = (tmplContext_t *)fbTemplateGetContext(subrec.tmpl);
                while ((subrec.rec = fbSTLNext(uint8_t, stl, subrec.rec))) {
                    reclen += idUpdateTemplateCounts(subtctx, &subrec);
                }
                fbSubTemplateListClear(stl);
            }
            break;
          case FB_SUB_TMPL_MULTI_LIST:
            {
                fbSubTemplateMultiList_t      *stml;
                fbSubTemplateMultiListEntry_t *entry;

                stml = ((fbSubTemplateMultiList_t *)
                        (record->rec + fbTemplateFieldGetOffset(field)));

                /* SMTL includes semantic(1) */
                reclen += 1;
                entry = NULL;
                while ((entry = fbSTMLNext(stml, entry))) {
                    /* STMLEntry includes tmpl-id(2) and length(2) */
                    reclen += 4;
                    subrec.tid =
                        fbSubTemplateMultiListEntryGetTemplateID(entry);
                    subrec.tmpl = fbSubTemplateMultiListEntryGetTemplate(entry);
                    subrec.rec = NULL;
                    subtctx =
                        (tmplContext_t *)fbTemplateGetContext(subrec.tmpl);
                    while ((subrec.rec =
                            fbSTMLEntryNext(uint8_t, entry, subrec.rec)))
                    {
                        reclen += idUpdateTemplateCounts(subtctx, &subrec);
                    }
                }
                fbSubTemplateMultiListClear(stml);
            }
            break;
          default:
            break;
        }
    }

    tmplContextUpdateReclenStats(tctx, reclen);

    return reclen;
}


static struct timespec *
idRecTimeToTimespec(
    struct timespec  *ts,
    uint64_t          rectime)
{
    ts->tv_sec = rectime / 1000000000;
    ts->tv_nsec = (rectime % 1000000000);
    return ts;
}


/* Returns a number representing whether or not the message contains
 * (option) templates sets, data sets or both */
static uint8_t
getMsgSetTypes(
    const uint8_t  *msgbuf,
    size_t          msg_len)
{
    uint16_t set_len;
    size_t   msg_offset = 16;    /* account for msg_header */
    uint16_t set_id;
    uint8_t  retval = 0;

    while (msg_offset < msg_len && retval != (TMPL_MSG | DATA_MSG)) {
        set_id = ntohs(*((uint16_t *)(msgbuf + msg_offset)));
        if (set_id == FB_TID_TS || set_id == FB_TID_OTS) {
            retval |= TMPL_MSG;
        } else if (set_id >= FB_TID_MIN_DATA) {
            retval |= DATA_MSG;
        }
        set_len = ntohs(*((uint16_t *)(msgbuf + msg_offset + 2)));
        msg_offset += set_len;
    }
    return retval;
}


/**
 *  Prints the template counts once all data has been read.
 */
static void
idPrintTemplateCounts(
    void)
{
    tmplContext_t *tctx;
    const char    *name;
    const char    *name_sep;
    uint32_t       avg;
    guint  i;

    /* either initialize `name` to the template-name header or set to the
     * empty string if no template names were seen */
    if (fbToolsTemplateNamesSeen()) {
        name = "Template-Name";
        name_sep = "| ";
    } else {
        name = "";
        name_sep = "";
    }

    g_ptr_array_sort(tmpl_ctx_array, idTemplateContextSort);

    fprintf(outfile,
            "     TemplateID| RecCount| MinLen| AvgLen| MaxLen%s%s\n",
            name_sep, name);
    for (i = 0; i < tmpl_ctx_array->len; ++i) {
        tctx = (tmplContext_t *)g_ptr_array_index(tmpl_ctx_array, i);
        if (name_sep[0]) {
            name = fbToolsTemplateNamesGetByTid(tctx->tid);
        }
        if (0 == tctx->rec_count) {
            if (!print_unused_counts) {
                continue;
            }

            /* leave sizes empty */
            fprintf(outfile, "%6d (%#06x)|%9u|%7s|%7s|%7s%s%s\n",
                    tctx->tid, tctx->tid, tctx->rec_count,
                    "", "", "", name_sep, name);
            continue;
        }
        if (tctx->is_fixedlen) {
            tctx->min_octets = tctx->max_octets = tctx->fixed_octets;
            avg = tctx->fixed_octets;
        } else {
            avg = tctx->all_octets / tctx->rec_count;
        }
        fprintf(outfile, "%6d (%#06x)|%9u|%7u|%7u|%7u%s%s\n",
                tctx->tid, tctx->tid, tctx->rec_count,
                tctx->min_octets, avg, tctx->max_octets, name_sep, name);
    }
}


/**
 *  Prints a summary of the message and record timestamps seen when
 *  --file-times is specified.
 */
static void
idPrintFileTimes(
    fileTimes_t  *file_times)
{
    struct timespec ts;
    gboolean no_times;

    /* if `min_rectime` was never changed, there is no record timestamp data */
    no_times = (UINT64_MAX == file_times->min_rectime);

    fprintf(outfile, "Time stats:\n");

    fprintf(outfile, "First record time:  ");
    if (no_times) {
        fprintf(outfile, "N/A\n");
    } else {
        idPrintTimestamp(
            outfile, idRecTimeToTimespec(&ts, file_times->first_rectime),
            FB_DT_MILSEC);
    }

    fprintf(outfile, "Last record time:   ");
    if (no_times) {
        fprintf(outfile, "N/A\n");
    } else {
        idPrintTimestamp(
            outfile, idRecTimeToTimespec(&ts, file_times->final_rectime),
            FB_DT_MILSEC);
    }

    fprintf(outfile, "Min record time:    ");
    if (no_times) {
        fprintf(outfile, "N/A\n");
    } else {
        idPrintTimestamp(
            outfile, idRecTimeToTimespec(&ts, file_times->min_rectime),
            FB_DT_MILSEC);
    }

    fprintf(outfile, "Max record time:    ");
    if (no_times) {
        fprintf(outfile, "N/A\n");
    } else {
        idPrintTimestamp(
            outfile, idRecTimeToTimespec(&ts, file_times->max_rectime),
            FB_DT_MILSEC);
    }

    fprintf(outfile, "First message time: ");
    if (0 == msg_count) {
        fprintf(outfile, "N/A\n");
    } else {
        ts.tv_nsec = 0;
        ts.tv_sec = file_times->start_msg_secs;
        idPrintTimestamp(outfile, &ts, FB_DT_SEC);
    }

    fprintf(outfile, "Last message time:  ");
    if (0 == msg_count) {
        fprintf(outfile, "N/A\n");
    } else {
        ts.tv_sec = file_times->last_msg_secs;
        idPrintTimestamp(outfile, &ts, FB_DT_SEC);
    }

    fprintf(outfile, "\n");
}


int
main(
    int    argc,
    char  *argv[])
{
    uint8_t        msgbuf[UINT16_MAX];
    fBuf_t        *fbuf = NULL;
    fbTemplate_t  *tmpl = NULL;
    GError        *err = NULL;
    FILE          *infile = NULL;
    gboolean       rc;
    fbSession_t   *session = NULL;
    tmplContext_t *tctx;
    int            rec_count = 0;
    size_t         got;
    fileTimes_t    file_times;
    fbRecord_t     record = FB_RECORD_INIT;

    idParseOptions(&argc, &argv);

    /* Initilize file_times */
    memset(&file_times, 0, sizeof(file_times));
    file_times.first_rectime = file_times.min_rectime = UINT64_MAX;

    /* Create an array to hold the tmplContext_t objects.  Objects are
     * inserted as they are created; if template-counts is requested, pointers
     * are sorted by TID prior to generating output. */
    tmpl_ctx_array = g_ptr_array_new_full(128, g_free);

    /* Create table to hold template names keyed by TID */
    fbToolsTemplateNamesInit();

    /* Open the input and output files */
    fbuf = fbToolsOpenInput(argv[0], &session, &infile);
    outfile = fbToolsOpenOutput();

    /* Disable automatic next-message mode */
    fBufSetAutomaticNextMessage(fbuf, FALSE);

    fbSessionAddNewTemplateCallback(session, idTemplateCallback, fbuf);

    /* Allocate buffer for a single in-memory record */
    record.reccapacity = RECBUF_CAPACITY_INITIAL;
    record.rec = g_new(uint8_t, record.reccapacity);

    /* Read until no more data or error */
    for (;;) {
        tmpl = fBufNextCollectionTemplate(fbuf, &record.tid, &err);
        if (!tmpl) {
            /* If no template - no message */
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOF)) {
                /* End of file */
                fBufFree(fbuf);
                g_clear_error(&err);
                eom = TRUE;
                break;
            }
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_BUFSZ)) {
                /* No current msg; attempt to read next message */
                got = fread(msgbuf, 1, 4, infile);
                if (got != 4) {
                    /* End of file */
                    if (0 != got) {
                        fprintf(stderr, "%s: Read %zu octets of 4 expected\n",
                                g_get_prgname(), got);
                    }
                    fBufFree(fbuf);
                    g_clear_error(&err);
                    eom = TRUE;
                    break;
                }
                /* Interpret first four octets as version and rec-length */
                if (ntohs(*((uint16_t *)(msgbuf))) != 10) {
                    /* Bad version */
                    fprintf(stderr,
                            "%s: Error: Illegal IPFIX Message version %#04x; "
                            "input is probably not an IPFIX Message stream.\n",
                            g_get_prgname(), ntohs(*((uint16_t *)(msgbuf))));
                    fBufFree(fbuf);
                    g_clear_error(&err);
                    eom = TRUE;
                    break;
                }
                msglen = ntohs(*((uint16_t *)(msgbuf + 2)));
                if (msglen < 16) {
                    /* Bad length */
                    fprintf(stderr,
                            "%s: Message length %zu too short to be IPFIX\n",
                            g_get_prgname(), msglen);
                    fBufFree(fbuf);
                    g_clear_error(&err);
                    eom = TRUE;
                    break;
                }
                /* Read remainder of message into the buffer */
                got = fread(msgbuf + 4, 1, msglen - 4, infile);
                if (got < msglen - 4) {
                    /* Failed to read message */
                    fprintf(stderr, "%s: Read %zu octets of %zu expected\n",
                            g_get_prgname(), got, msglen - 4);
                    fBufFree(fbuf);
                    g_clear_error(&err);
                    eom = TRUE;
                    break;
                }
                /* Get values from IPFIX header */
                sequence_number = ntohl(*((uint32_t *)(msgbuf + 8)));
                curr_msg_set_types = getMsgSetTypes(msgbuf, msglen);
                /* Hand the message buffer to fixbuf */
                fBufSetBuffer(fbuf, msgbuf, msglen);
                offset += msglen;
                eom = TRUE;
            } else if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOM)) {
                eom = TRUE;
            } else {
                fprintf(stderr, "%s: Warning: %s\n",
                        g_get_prgname(), err->message);
            }
            g_clear_error(&err);
            continue;
        }

        if (eom) {
            idNewMessage(fbuf);

            if (print_file_times) {
                file_times.last_msg_secs = fBufGetExportTime(fbuf);
                if (file_times.start_msg_secs == 0) {
                    file_times.start_msg_secs = file_times.last_msg_secs;
                }
            }
        }

        tctx = fbTemplateGetContext(tmpl);
        record.recsize = tctx->len;

        if (record.recsize > record.reccapacity) {
            do {
                record.reccapacity <<= 1;
            } while (record.recsize > record.reccapacity);
            record.rec = g_renew(uint8_t, record.rec, record.reccapacity);
        }
        memset(record.rec, 0, record.reccapacity);

        /* Get the next record */
        rc = fBufNextRecord(fbuf, &record, &err);
        if (FALSE == rc) {
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOF)) {
                eom = TRUE;
                fprintf(stderr, "%s: END OF FILE\n", g_get_prgname());
                fBufFree(fbuf);
                g_clear_error(&err);
                break;
            }
            if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_EOM)) {
                eom = TRUE;
            } else if (g_error_matches(err, FB_ERROR_DOMAIN, FB_ERROR_BUFSZ)) {
                eom = TRUE;
            } else {
                fprintf(stderr, "%s: Warning: %s\n",
                        g_get_prgname(), err->message);
            }
            g_clear_error(&err);
            continue;
        }

        ++rec_count;
        ++msg_rec_count;

        if (tctx->meta_template) {
            fbToolsTemplateNamesExamineRecord(tctx->meta_template, &record);
        } else if (tctx->meta_element) {
            fbToolsRfc5610ExamineRecord(tctx->meta_element, &record);
        }

        /* Gather timestamps for file stats */
        if (print_file_times && tctx->contains_end_time) {
            fbRecordValue_t  value;
            fbRecordGetValueForField(&record, &tctx->end_time_index, &value);
            file_times.final_rectime = (value.v.dt.tv_sec * 1000000000
                                        + value.v.dt.tv_nsec);
            if (file_times.first_rectime == UINT64_MAX) {
                file_times.first_rectime =
                  file_times.min_rectime =
                  file_times.max_rectime = file_times.final_rectime;
            } else if (file_times.final_rectime > file_times.max_rectime) {
                file_times.max_rectime = file_times.final_rectime;
            } else if (file_times.final_rectime < file_times.min_rectime) {
                file_times.min_rectime = file_times.final_rectime;
            }
        }

        if (print_data) {
            idPrintDataRecord(outfile, tctx, &record, rec_count,
                              "Data", "");
        } else if (print_template_counts) {
            idUpdateTemplateCounts(tctx, &record);
        } else {
            fbRecordFreeLists(&record);
        }
    }

    if (eom) {
        idCloseMessage();
    }

    g_free(record.rec);

    /* account for missing normally newline printed by idCloseMessage */
    if (suppress_messages && (print_data || print_templates)) {
        fprintf(outfile, "\n");
    }

    if (print_template_counts) {
        idPrintTemplateCounts();
        if (print_data || print_templates || print_messages
            || print_file_times)
        {
            fprintf(outfile, "\n");
        }
    }
    if (print_file_times) {
        idPrintFileTimes(&file_times);
    }

    if (print_data || print_templates || print_messages || print_file_times) {
        fprintf(outfile, "*** File Stats: %u Messages, %u Data Records, "
                "%u Template Records ***\n", msg_count, rec_count, tmpl_count);
    }

    g_ptr_array_free(tmpl_ctx_array, TRUE);

    fbToolsTemplateNamesDestroy();

    return 0;
}
